from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Dict, Tuple, Set, List, Union

from data_to_paper.base_products.file_descriptions import DataFileDescriptions, DataFileDescription
from data_to_paper.base_steps import LiteratureSearch
from data_to_paper.base_steps.literature_search import LiteratureSearchParams
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.code_and_output_files.ref_numeric_values import replace_hyperlinks_with_values
from data_to_paper.code_and_output_files.referencable_text import hypertarget_if_referencable_text_product
from data_to_paper.conversation.stage import Stage
from data_to_paper.latex import extract_latex_section_from_response
from data_to_paper.latex.latex_to_pdf import evaluate_latex_num_command
from data_to_paper.latex.tables import add_displayitems_to_paper_section, get_displayitem_caption

from data_to_paper.research_types.hypothesis_testing.cast import ScientificAgent
from data_to_paper.research_types.hypothesis_testing.product_types import HypothesisTestingPlanProduct, \
    NoveltyAssessmentProduct, GoalAndHypothesisProduct, MostSimilarPapersProduct, NoveltySummaryProduct
from data_to_paper.research_types.hypothesis_testing.scientific_stage import ScientificStage, \
    SECTION_NAMES_TO_WRITING_STAGES
from data_to_paper.code_and_output_files.code_and_output import CodeAndOutput

from data_to_paper.utils.nice_list import NiceList
from data_to_paper.base_products import Products, NameDescriptionStageGenerator, ProductGenerator
from data_to_paper.utils.types import ListBasedSet, MemoryDict
from data_to_paper.servers.custom_types import Citation

CODE_STEPS_TO_STAGES_NAMES_AGENTS: Dict[str, Tuple[Stage, str, ScientificAgent]] = {
    'data_exploration': (ScientificStage.EXPLORATION, 'Data Exploration', ScientificAgent.DataExplorer),
    # 'data_preprocessing': (ScientificStage.PREPROCESSING, 'Data Preprocessing', ScientificAgent.DataPreprocessor),
    'data_analysis': (ScientificStage.CODE, 'Data Analysis', ScientificAgent.Debugger),
    'data_to_latex': (ScientificStage.DISPLAYITEMS, 'LaTeX Table Design', ScientificAgent.InterpretationReviewer),
}


class HypertargetPrefix(Enum):
    """
    Prefixes for hypertargets.
    """
    GENERAL_FILE_DESCRIPTION = 'S'
    FILE_DESCRIPTIONS = ('T', 'U', 'V', 'W', 'X', 'Y', 'Z')
    ADDITIONAL_RESULTS = ('R',)
    LATEX_TABLES = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H')


def get_code_stage(code_step: str) -> Stage:
    """
    Return the stage of the code step.
    """
    return CODE_STEPS_TO_STAGES_NAMES_AGENTS[code_step][0]


def get_code_name(code_step: str) -> str:
    """
    Return the name of the code step.
    """
    return CODE_STEPS_TO_STAGES_NAMES_AGENTS[code_step][1]


def get_code_agent(code_step: str) -> ScientificAgent:
    """
    Return the agent of the code step.
    """
    return CODE_STEPS_TO_STAGES_NAMES_AGENTS[code_step][2]


def convert_description_of_created_files_to_string(description_of_created_files: Dict[str, str]) -> Optional[str]:
    """
    Convert the description of created files to a string.
    """
    if not description_of_created_files:
        return None
    return '\n'.join(
        f'File "{file_name}":\n\n{file_description}'
        for file_name, file_description in description_of_created_files.items()
    )


STAGE_AND_SCOPE_TO_LITERATURE_SEARCH_PARAMS = {
    'goal': ({
        'dataset': LiteratureSearchParams(12, 2, 2.0, False),
        'questions': LiteratureSearchParams(12, 2, 2.0, False),
    }, "Literature Search for Goal", ScientificStage.LITERATURE_REVIEW_GOAL),
    'writing': ({
        'background': LiteratureSearchParams(12, 5, 2.0, True),
        'dataset': LiteratureSearchParams(12, 2, 2.0, False),
        'methods': LiteratureSearchParams(6, 10, 1.5, False),
        'results': LiteratureSearchParams(12, 1, 2.0, True),
    }, "Literature Search for Writing", ScientificStage.LITERATURE_REVIEW_WRITING),
}


def _create_literature_search(stage: str) -> LiteratureSearch:
    params = STAGE_AND_SCOPE_TO_LITERATURE_SEARCH_PARAMS[stage]
    return LiteratureSearch(name=params[1], stage=params[2], scopes_to_search_params=params[0])


def _get_literature_searchs() -> Dict[str, LiteratureSearch]:
    return {stage: _create_literature_search(stage) for stage in STAGE_AND_SCOPE_TO_LITERATURE_SEARCH_PARAMS}


@dataclass
class ScientificProducts(Products):
    """
    All products of the scientific process.
    """
    data_file_descriptions: DataFileDescriptions = field(default_factory=DataFileDescriptions)
    codes_and_outputs: Dict[str, CodeAndOutput] = field(default_factory=dict)
    research_goal: GoalAndHypothesisProduct = None
    novelty_assessment: NoveltyAssessmentProduct = None
    literature_search: Dict[str, LiteratureSearch] = field(default_factory=_get_literature_searchs)
    most_similar_papers: MostSimilarPapersProduct = None
    hypothesis_testing_plan: HypothesisTestingPlanProduct = None
    paper_sections_and_optional_citations: Dict[str, Union[str, Tuple[str, Set[Citation]]]] = \
        field(default_factory=MemoryDict)

    def get_created_dfs(self) -> List[str]:
        return [file for file in self.codes_and_outputs['data_analysis'].created_files.get_created_content_files()
                if file.startswith('tbl_') or file.startswith('fig_')]

    def get_number_of_created_dfs(self) -> int:
        return len(self.get_created_dfs())

    def get_latex_displayitems(self, view_purpose: ViewPurpose = ViewPurpose.PRODUCT) -> Dict[str, List[str]]:
        """
        Return the tables.
        """
        return {'results': [
            content for file, content
            in self.codes_and_outputs[
                'data_to_latex'].created_files.get_created_content_files_to_pretty_contents(view_purpose).items()
            if file.endswith('.pkl')]}

    def get_all_latex_tables(self, view_purpose: ViewPurpose) -> List[str]:
        """
        Return the tables from all sections.
        """
        return [table for tables in self.get_latex_displayitems(view_purpose).values() for table in tables]

    @property
    def all_file_descriptions(self) -> DataFileDescriptions:
        """
        Return the description of all files.
        """
        desc = DataFileDescriptions.from_other(self.data_file_descriptions)
        for code_and_output in self.codes_and_outputs.values():
            if code_and_output.description_of_created_files is not None:
                desc += code_and_output.description_of_created_files
            else:
                desc += [DataFileDescription(file_path=created_file)
                         for created_file in code_and_output.created_files.get_created_data_files()]
        desc.data_folder = self.data_file_descriptions.data_folder
        return desc

    def get_file_headers(self, code_step: str):
        """
        Return the file headers of a given code_step.
        """
        code_and_output = self.codes_and_outputs[code_step]
        created_files = code_and_output.created_files.get_created_data_files()
        if not created_files:
            return None
        return DataFileDescriptions(
            [DataFileDescription(file_path=created_file) for created_file in created_files],
            data_folder=self.data_file_descriptions.data_folder)

    def get_paper_sections_and_citations(self, remove_hyperlinks: bool = False,
                                         format_num_command: Optional[bool] = False,
                                         ) -> Dict[str, Tuple[str, Set[Citation]]]:
        """
        Like paper_sections_and_optional_citations, but adding empty citations to sections without citations.
        remove_hyperlinks: whether to remove the hyperlinks in the latex sections.
        format_num_command: whether to format the \num{} commands in the latex sections:
            - True, replace the \num{} commands with their values.
            - None, keep the \num{} commands, but remove the explanation.
            - False, keep the \num{} commands.
            format_num_command is only effective with remove_hyperlinks=True.
        """
        section_names_to_sections_and_citations = {}
        for section_name, section_and_optional_citations in self.paper_sections_and_optional_citations.items():
            if isinstance(section_and_optional_citations, str):
                section = section_and_optional_citations
                citations = set()
            else:
                section, citations = section_and_optional_citations
            if remove_hyperlinks:
                section = replace_hyperlinks_with_values(section)
                if format_num_command is not False:
                    section = evaluate_latex_num_command(section, just_strip_explanation=format_num_command is None)[0]

            section_names_to_sections_and_citations[section_name] = (section, citations)
        return section_names_to_sections_and_citations

    def get_paper_sections_without_citations(self, remove_hyperlinks: bool = False,
                                             format_num_command: Optional[bool] = False
                                             ) -> Dict[str, str]:
        return {section_name: section
                for section_name, (section, citation) in self.get_paper_sections_and_citations(
                    remove_hyperlinks=remove_hyperlinks, format_num_command=format_num_command).items()}

    def get_all_cited_citations(self) -> NiceList[Citation]:
        """
        Return the citations of the paper.
        """
        citations = ListBasedSet()
        for section_content, section_citations in self.get_paper_sections_and_citations().values():
            citations.update(section_citations)
        return NiceList(citations, separator='\n\n')

    def get_tabled_paper_sections(self, view_purpose: ViewPurpose) -> Dict[str, str]:
        """
        Return the paper sections with tables inserted at the right places.
        """
        latex_displayitems = self.get_latex_displayitems(view_purpose)
        return {section_name: section if section_name not in latex_displayitems
                else add_displayitems_to_paper_section(section, latex_displayitems[section_name])
                for section_name, section in self.get_paper_sections_without_citations().items()}

    def get_title(self) -> str:
        """
        Return the title of the paper.
        """
        latex = self.get_paper_sections_without_citations()['title']
        return extract_latex_section_from_response(latex, 'title', keep_tags=False)

    def get_abstract(self) -> str:
        """
        Return the abstract of the paper.
        """
        latex = self.get_paper_sections_without_citations()['abstract']
        return extract_latex_section_from_response(latex, 'abstract', keep_tags=False)

    def _get_generators(self) -> Dict[str, NameDescriptionStageGenerator]:
        return {
            **super()._get_generators(),

            # DATA
            # ====

            'general_dataset_description': NameDescriptionStageGenerator(
                'Overall Description of the Dataset',
                '{}',
                ScientificStage.DATA,
                lambda: hypertarget_if_referencable_text_product(self.data_file_descriptions.general_description,
                                                                 ViewPurpose.PRODUCT, level=2),
            ),

            'data_file_descriptions': NameDescriptionStageGenerator(
                'Description of the Original Dataset',
                '{}',
                ScientificStage.DATA,
                lambda: self.data_file_descriptions,
            ),

            'data_file_descriptions_no_headers': NameDescriptionStageGenerator(
                'Description of the Original Dataset',
                '{}',
                ScientificStage.DATA,
                lambda: self.data_file_descriptions.pretty_repr(num_lines=0),
            ),

            'data_file_descriptions_no_headers_linked': NameDescriptionStageGenerator(
                'Description of the Original Dataset (with hypertargets)',
                '{}',
                ScientificStage.DATA,
                lambda: self.data_file_descriptions.pretty_repr(num_lines=0,
                                                                view_purpose=ViewPurpose.HYPERTARGET_PRODUCT),
            ),

            'all_file_descriptions': NameDescriptionStageGenerator(
                'Description of the Dataset',
                '{}',
                ScientificStage.DATA,
                lambda: self.all_file_descriptions,
            ),

            # GOAL AND PLAN
            # ==============

            'research_goal': ProductGenerator(
                lambda: self.research_goal,
                {},
            ),

            'hypothesis_testing_plan': ProductGenerator(
                lambda: self.hypothesis_testing_plan,
                {},
            ),

            # LITERATURE SEARCH
            # =================

            'literature_search:{}': ProductGenerator(
                lambda stage: self.literature_search[stage],
                lambda stage: dict(stage=stage, scope=None),
            ),

            'literature_search:{}:{}': ProductGenerator(
                lambda stage, scope: self.literature_search[stage],
                lambda stage, scope: dict(stage=stage, scope=scope),
            ),

            'most_similar_papers': ProductGenerator(
                lambda: self.most_similar_papers,
                {},
            ),

            'novelty_assessment': ProductGenerator(
                lambda: NoveltySummaryProduct(novelty_assessment=self.novelty_assessment,
                                              most_similar_papers=self.most_similar_papers),
                {},
            ),

            # CODE
            # ====

            'codes:{}': NameDescriptionStageGenerator(
                '{code_name} Code',
                'Here is our {code_name} Code:\n```python\n{code}\n```\n',
                lambda code_step: get_code_stage(code_step),
                lambda code_step: {'code': self.codes_and_outputs[code_step].code,
                                   'code_name': self.codes_and_outputs[code_step].name},
            ),

            'outputs:{}': NameDescriptionStageGenerator(
                'Output of the {code_name} Code',
                'Here is the Output of our {code_name} code:\n{output}',
                lambda code_step: get_code_stage(code_step),
                lambda code_step: {
                    'output': self.codes_and_outputs[code_step].created_files.
                    get_created_content_files_and_contents_as_single_str(view_purpose=ViewPurpose.PRODUCT,
                                                                         header_level=3),
                    'code_name': self.codes_and_outputs[code_step].name},
            ),

            'code_explanation:{}': NameDescriptionStageGenerator(
                '{code_name} Code Description',
                'Here is an explanation of our {code_name} code:\n\n{code_explanation}',
                lambda code_step: get_code_stage(code_step),
                lambda code_step: {
                    'code_name': self.codes_and_outputs[code_step].name,
                    'code_explanation': self.codes_and_outputs[code_step].code_explanation},
            ),

            'codes_and_outputs:{}': NameDescriptionStageGenerator(
                '{code_name} Code and Output',
                '{code_description}\n\n{output_description}',
                lambda code_step: get_code_stage(code_step),
                lambda code_step: {
                    'code_name': self.codes_and_outputs[code_step].name,
                    'code_description': self.get_description('codes:' + code_step),
                    'output_description': self.get_description('outputs:' + code_step)},
            ),

            'codes_and_outputs_with_explanations:{}': NameDescriptionStageGenerator(
                '{code_name} Code and Output',
                '{description}',
                lambda code_step: get_code_stage(code_step),
                lambda code_step: {
                    'code_name': self.codes_and_outputs[code_step].name,
                    'description': self.codes_and_outputs[code_step].to_text(with_header=False)},
            ),

            'created_files_content:{}:{}': NameDescriptionStageGenerator(
                'Content of Files Created by the {code_name} Code',
                'Here is the content of {which_files} created by the {code_name} code:\n\n{created_files_content}',
                lambda code_step, filespec: get_code_stage(code_step),
                lambda code_step, filespec: {
                    'created_files_content':
                        self.codes_and_outputs[
                            code_step].created_files.get_created_content_files_and_contents_as_single_str(
                            view_purpose=ViewPurpose.CODE_REVIEW, match_filename=filespec, header_level=3),
                    'which_files': 'all files' if filespec == '*' else f'files "{filespec}"',
                    'code_name': self.codes_and_outputs[code_step].name},
            ),

            'created_files_description:{}': NameDescriptionStageGenerator(
                'Description of Files Created by the {code_name} Code',
                'We can use these files created by the {code_name} code:\n\n{created_files_description}',
                lambda code_step: get_code_stage(code_step),
                lambda code_step: {
                    'created_files_description': DataFileDescriptions(
                        self.data_file_descriptions + self.codes_and_outputs[code_step].description_of_created_files,
                        data_folder=self.codes_and_outputs[code_step].description_of_created_files.data_folder)
                    if self.codes_and_outputs[code_step].description_of_created_files is not None else None,
                    'code_name': self.codes_and_outputs[code_step].name},
            ),

            'created_files_headers:{}': NameDescriptionStageGenerator(
                'Headers of Files Created by the {code_name} Code',
                'Here are the headers of the files created by the {code_name} code:\n\n{created_files_headers}',
                lambda code_step: get_code_stage(code_step),
                lambda code_step: {
                    'created_files_headers': self.get_file_headers(code_step),
                    'code_name': self.codes_and_outputs[code_step].name},
            ),

            # WRITING
            # =======

            'title_and_abstract_first': NameDescriptionStageGenerator(
                'Title and Abstract (initial draft)',
                "```latex\n{}\n\n{}```",
                ScientificStage.INTERPRETATION,
                lambda: (self.get_paper_sections_without_citations()['title'],
                         self.get_paper_sections_without_citations()['abstract']),
            ),

            'title_and_abstract': NameDescriptionStageGenerator(
                'Title and Abstract',
                "```latex\n{}\n\n{}```",
                ScientificStage.WRITING_TITLE_AND_ABSTRACT,
                lambda: (self.get_paper_sections_without_citations()['title'],
                         self.get_paper_sections_without_citations()['abstract']),
            ),

            'paper_sections:{}': NameDescriptionStageGenerator(
                '{section_name} Section of the Paper',
                '```latex\n{content}\n```',
                lambda section_name: SECTION_NAMES_TO_WRITING_STAGES[section_name],
                lambda section_name: {'section_name': section_name.title(),
                                      'content': self.get_paper_sections_without_citations(
                                          remove_hyperlinks=True, format_num_command=None)[section_name],
                                      },
            ),

            'latex_displayitems': NameDescriptionStageGenerator(
                'Displayitems of the Paper',
                'Here are the displayitems created by our data analysis code '
                '(figure/table latex representations of the df_?.pkl dataframes):\n\n{}',
                ScientificStage.DISPLAYITEMS,
                lambda: None if not self.get_all_latex_tables(ViewPurpose.PRODUCT) else
                '\n\n'.join([f'- "{get_displayitem_caption(table, first_line_only=True)}":\n\n'
                             f'{table}'
                             for table in self.get_all_latex_tables(ViewPurpose.PRODUCT)]),
            ),

            'latex_displayitems_linked': NameDescriptionStageGenerator(
                'Displayitems of the Paper with hypertargets',
                'Here are the displayitems created by our data analysis code '
                '(figure/table latex representations of the df_?.pkl dataframes, with hypertargets):\n\n{}',
                ScientificStage.DISPLAYITEMS,
                lambda: None if not self.get_all_latex_tables(ViewPurpose.HYPERTARGET_PRODUCT) else
                '\n\n'.join([f'- "{get_displayitem_caption(table, first_line_only=True)}":\n\n'
                             f'{table}'
                             for table in self.get_all_latex_tables(ViewPurpose.HYPERTARGET_PRODUCT)]),
            ),

            'additional_results': NameDescriptionStageGenerator(
                'Additional Results (additional_results.pkl)',
                'Here are some additional numeric values that may be helpful in writing the paper:\n\n{}',
                ScientificStage.INTERPRETATION,
                lambda: self.codes_and_outputs[
                    'data_analysis'].created_files.get_created_content_files_to_pretty_contents(
                    view_purpose=ViewPurpose.PRODUCT, header_level=3)['additional_results.pkl'],
            ),

            'additional_results_linked': NameDescriptionStageGenerator(
                'Additional Results (additional_results.pkl) with hypertargets',
                'Here are some additional numeric values that may be helpful in writing the paper:\n\n{}',
                ScientificStage.INTERPRETATION,
                lambda: self.codes_and_outputs[
                    'data_analysis'].created_files.get_created_content_files_to_pretty_contents(
                    view_purpose=ViewPurpose.HYPERTARGET_PRODUCT, header_level=3)['additional_results.pkl'],
            ),
        }



import functools
import inspect
from dataclasses import dataclass
from typing import Iterable

from data_to_paper.run_gpt_code.base_run_contexts import MultiRunContext
from data_to_paper.run_gpt_code.attr_replacers import SystematicMethodReplacerContext, \
    PreventAssignmentToAttrs, AttrReplacer
from data_to_paper.run_gpt_code.run_issues import CodeProblem, RunIssue
from data_to_paper.text import dedent_triple_quote_str

from ..pvalue import convert_to_p_value, TrackPValueCreationFuncs


@dataclass
class SklearnFitOverride(SystematicMethodReplacerContext):

    def _get_all_modules(self) -> list:
        # add here all modules that have classes with fit methods:
        from sklearn import linear_model, svm
        return [linear_model, svm]

    def _should_replace(self, parent, attr_name, attr) -> bool:
        return attr_name.startswith('fit')

    def _get_custom_wrapper(self, parent, attr_name, original_func):

        @functools.wraps(original_func)
        def wrapped(obj, *args, **kwargs):
            result = original_func(obj, *args, **kwargs)
            if self._is_called_from_user_script():
                if hasattr(obj, '_prior_fit_results') and obj._prior_fit_results is result:
                    raise RunIssue.from_current_tb(
                        category='Sklearn: good practices',
                        issue=f"The `{original_func.__name__}` function was already called on this object. ",
                        instructions=f"Multiple calls should be avoided as the same result instance is returned again.",
                        code_problem=CodeProblem.RuntimeError,
                    )
                obj._prior_fit_results = result

            return result

        return wrapped


def _f_regression(*args, original_func=None, **kwargs):
    p = original_func(*args, **kwargs)
    return (p[0], convert_to_p_value(p[1], created_by='f_regression'))


@dataclass
class SklearnPValue(AttrReplacer, TrackPValueCreationFuncs):
    """
    f_regression should return array of PValue objects instead of floats.
    """
    obj_import_str: str = 'sklearn.feature_selection'
    attr: str = 'f_regression'
    wrapper: callable = _f_regression
    send_original_to_wrapper: bool = True


@dataclass
class SklearnSearchLimitCheck(SystematicMethodReplacerContext):
    max_iterations: int = 12  # Default max iterations limit

    def _get_all_parents(self) -> list:
        # add here all modules that have classes with fit methods:
        from sklearn.model_selection import ParameterGrid, ParameterSampler
        return [ParameterGrid, ParameterSampler]

    def _should_replace(self, parent, attr_name, attr) -> bool:
        return attr_name == "__len__"

    def _get_estimator_class_name(self):
        # Inspect the stack and find the class name of the estimator
        for frame_record in inspect.stack():
            if 'self' in frame_record.frame.f_locals:
                instance = frame_record.frame.f_locals['self']
                if hasattr(instance, 'fit'):
                    return instance.estimator.__class__.__name__
        return 'Unknown'

    def _get_custom_wrapper(self, parent, attr_name, original_func):

        @functools.wraps(original_func)
        def wrapped(obj, *args, **kwargs):
            original_len = original_func(obj, *args, **kwargs)
            if original_len > self.max_iterations:
                estimator_class_name = self._get_estimator_class_name()
                raise RunIssue.from_current_tb(
                    category='Sklearn: too many iterations',
                    issue=dedent_triple_quote_str(f"""
                        The presumed total number of training iterations ({original_len}) for the estimator \t
                        {estimator_class_name} exceeds the maximum allowed iterations ({self.max_iterations}).
                        Notice that the amount of iterations is a multiplication of the numbers of possible values \t
                        for each parameter when using GridSearchCV or n_iter when using RandomizedSearchCV.
                        """),
                    instructions=f"use only a subset of the parameters or reduce the number of iterations.",
                    code_problem=CodeProblem.RuntimeError)
            return original_len

        return wrapped


@dataclass
class SklearnRandomStateOverride(SystematicMethodReplacerContext):
    """
    This class overrides any provided sklearn class constructor (__init__) that supports random_state
    from sklearn to set random_state=0, ensuring reproducibility.
    """
    def _get_all_parents(self) -> list:
        # Add here all classes that have a random_state parameter in their constructor:
        from sklearn.ensemble import RandomForestRegressor
        from sklearn.linear_model import ElasticNet
        from sklearn.neural_network import MLPRegressor
        return [RandomForestRegressor, ElasticNet, MLPRegressor]

    def _should_replace(self, parent, attr_name, attr) -> bool:
        if not attr_name == '__init__':
            return False
        sig = inspect.signature(attr)
        assert 'random_state' in sig.parameters, f"Expected random_state to be in the signature of {parent.__name__}"
        return True

    def _get_custom_wrapper(self, parent, attr_name, original_func):

        @functools.wraps(original_func)
        def wrapped(obj, *args, **kwargs):
            # Set random_state=0 if not explicitly passed
            if 'random_state' not in kwargs:
                kwargs['random_state'] = 0

            # Call the original constructor
            return original_func(obj, *args, **kwargs)

        return wrapped


@dataclass
class SklearnSingleNNSizeOverride(PreventAssignmentToAttrs):
    """
    This class overrides MLPRegressor and MLPClassifier to raise warning if the hidden_layer_sizes is too large.
    """
    obj_import_str: str = None  # 'sklearn.neural_network.MLPRegressor', 'sklearn.neural_network.MLPClassifier'
    forbidden_set_attrs: Iterable[str] = ('hidden_layer_sizes', )
    max_layers: int = 2
    max_neurons_per_layer: int = 50

    def _is_called_from_user_script(self) -> bool:
        return True

    def _raise_exception(self, attr, value):
        # if value is the same as default, do not raise exception - get value using inspect and compare
        if value == inspect.signature(self.obj).parameters[attr].default:
            return
        # Check depth:
        instructions = f"Please use a smaller hidden_layer_sizes of up to {self.max_layers} layers with " \
                       f"maximum of {self.max_neurons_per_layer} neurons per layer."
        if len(value) > self.max_layers:
            raise RunIssue.from_current_tb(
                category='Too many layers',
                issue=f"The hidden_layer_sizes ({len(value)}) is too large!\n",
                instructions=instructions,
                code_problem=CodeProblem.RuntimeError)
        # Check width:
        for layer, layer_size in enumerate(value):
            if layer_size > self.max_neurons_per_layer:
                raise RunIssue.from_current_tb(
                    category='Too large hidden layer',
                    issue=f"The hidden_layer_sizes, has a layer ({layer}) with too many neurons ({layer_size})!\n",
                    instructions=instructions,
                    code_problem=CodeProblem.RuntimeError)


@dataclass
class SklearnNNSizeOverride(MultiRunContext):
    contexts: Iterable[SklearnSingleNNSizeOverride] = (
        SklearnSingleNNSizeOverride(obj_import_str='sklearn.neural_network.MLPRegressor'),
        SklearnSingleNNSizeOverride(obj_import_str='sklearn.neural_network.MLPClassifier'),
    )



from dataclasses import dataclass

from typing import Union

from data_to_paper.interactive.app_interactor import AppInteractor
from data_to_paper.types import HumanReviewType
from data_to_paper.env import DEFAULT_HUMAN_REVIEW_TYPE


@dataclass
class HumanReviewAppInteractor(AppInteractor):
    human_review: Union[bool, HumanReviewType] = True
    # whether to ask for human review
    # True: HumanReviewType = DEFAULT_HUMAN_REVIEW_TYPE
    # False: HumanReviewType = HumanReviewType.NONE

    @property
    def actual_human_review(self):
        if self.app is None:
            return HumanReviewType.NONE
        if isinstance(self.human_review, bool):
            return DEFAULT_HUMAN_REVIEW_TYPE if self.human_review else HumanReviewType.NONE
        return self.human_review



import re
from typing import Optional, Dict

import pandas as pd
from pathlib import Path

from data_to_paper.latex.clean_latex import process_latex_text_and_math
from data_to_paper.run_gpt_code.overrides.dataframes.utils import df_to_html_with_value_format
from data_to_paper.run_gpt_code.overrides.pvalue import pvalue_on_str_for_latex
from data_to_paper.utils.check_type import raise_on_wrong_func_argument_types_decorator

from .describe import df_to_numerically_labeled_latex
from .note_and_legend import convert_note_and_glossary_to_latex_table_caption, \
    convert_note_and_glossary_to_html
from .utils import convert_filename_to_label

THREEPARTTABLE = r"""\begin{table}[htbp]
\centering
\begin{threeparttable}
<caption>
<label>
<tabular>
\begin{tablenotes}
<note_and_glossary>
\end{tablenotes}
\end{threeparttable}
\end{table}
"""

THREEPARTTABLE_WIDE = r"""\begin{table}[h]
<caption>
<label>
\begin{threeparttable}
\renewcommand{\TPTminimum}{\linewidth}
\makebox[\linewidth]{%
<tabular>}
\begin{tablenotes}
\footnotesize
<note_and_glossary>
\end{tablenotes}
\end{threeparttable}
\end{table}
"""


HTML_TABLE_WITH_LABEL_AND_CAPTION = r"""
<b>{caption}</b>
<style>
    thead th {
        text-align: left;  /* Align the column headers to the left */
    }
    tbody th {
        text-align: left;  /* Align the row headers (index) to the left */
    }
</style>
{table}
{note_and_glossary}
"""


@raise_on_wrong_func_argument_types_decorator
def df_to_latex(df: pd.DataFrame, filename: Optional[str],
                caption: Optional[str] = None,
                label: Optional[str] = None,
                note: Optional[str] = None,
                glossary: Optional[Dict[str, str]] = None,
                is_wide: bool = True,
                is_html: Optional[bool] = False,
                figure_folder: Optional[Path] = None,  # needed for compatibility with df_to_figure
                should_format: bool = False,
                **kwargs):
    """
    Create a latex table with a note.
    Same as df.to_latex, but with a note and glossary.
    is_html: If True, return an HTML table instead of a LaTeX table.
        if False, return a LaTeX table.
        if None, do nothing. (used to raise on wrong argument types)
    """
    label = convert_filename_to_label(filename, label)
    label = 'table:' + label

    index = kwargs.get('index', True)
    if is_html is True:
        html_caption = caption if caption else ''
        regular_html_table = df_to_html_with_value_format(df, border=0, justify='left', **kwargs)
        note_and_glossary_html = convert_note_and_glossary_to_html(df, note, glossary, index)
        return HTML_TABLE_WITH_LABEL_AND_CAPTION.replace('{caption}', html_caption) \
            .replace('{table}', regular_html_table) \
            .replace('{note_and_glossary}', note_and_glossary_html)

    if is_html is False:
        with pvalue_on_str_for_latex():
            regular_latex_table = df_to_numerically_labeled_latex(df, should_format=should_format, **kwargs)
        tabular_part = get_tabular_block(regular_latex_table)
        latex_caption = r'\caption{' + process_latex_text_and_math(caption) + '}\n' if caption else ''
        label = r'\label{' + label + '}\n' if label else ''

        note_and_glossary = convert_note_and_glossary_to_latex_table_caption(df, note, glossary, index)

        template = THREEPARTTABLE if not is_wide else THREEPARTTABLE_WIDE
        latex = template.replace('<tabular>', tabular_part) \
            .replace('<caption>\n', latex_caption) \
            .replace('<label>\n', label) \
            .replace('<note_and_glossary>', note_and_glossary)

        return latex


def get_tabular_block(latex_table: str) -> str:
    """
    Extract the tabular block of the table.
    """
    return re.search(pattern=r'\\begin{tabular}.*\n(.*)\\end{tabular}', string=latex_table, flags=re.DOTALL).group(0)



from __future__ import annotations

import builtins
import os
import tempfile
import warnings

from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Any, Iterable, Callable, List, Type, Dict, Optional, Union, Tuple, NamedTuple

from pathlib import Path

from data_to_paper.utils.file_utils import is_name_matches_list_of_wildcard_names
from data_to_paper.utils.types import ListBasedSet
from data_to_paper.text import dedent_triple_quote_str

from .exceptions import CodeWriteForbiddenFile, CodeReadForbiddenFile, \
    CodeImportForbiddenModule, UnAllowedFilesCreated
from .run_issues import CodeProblem, RunIssue
from data_to_paper.code_and_output_files.output_file_requirements import OutputFileRequirements
from .base_run_contexts import SingletonRegisteredRunContext


@dataclass
class ProvideData(SingletonRegisteredRunContext):
    data: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def get_item(cls, key: str) -> Any:
        self = cls.get_runtime_instance()
        return self.data[key]


@dataclass
class IssueCollector(SingletonRegisteredRunContext):
    pass


def _is_temp_file(file_name):
    return file_name.startswith(tempfile.gettempdir())


@dataclass
class PreventFileOpen(SingletonRegisteredRunContext):
    SYSTEM_FILES = ['templates/latex_table.tpl', 'templates/latex_longtable.tpl', 'ttf/DejaVuSans.ttf',
                    'LICENSE.txt', 'LICENSE', 'LICENSE.md']
    SYSTEM_FOLDERS = \
        [r'C:\Windows', r'C:\Program Files', r'C:\Program Files (x86)'] if os.name == 'nt' \
        else ['/usr', '/etc', '/bin', '/sbin', '/sys', '/dev', '/var', '/opt', '/proc']
    allowed_read_files: Iterable[str] = 'all'  # list of wildcard names,  'all' means allow all, [] means allow none
    allowed_write_files: Iterable[str] = 'all'  # list of wildcard names,  'all' means allow all, [] means allow none
    allowed_write_folder: Optional[str] = None
    original_open: Optional[Callable] = None

    def _reversible_enter(self):
        self.original_open = builtins.open
        builtins.open = self.open_wrapper
        return super()._reversible_enter()

    def _reversible_exit(self):
        builtins.open = self.original_open
        self.original_open = None
        return super()._reversible_exit()

    def is_allowed_read_file(self, file_name: str) -> bool:
        return self.allowed_read_files == 'all' or \
            is_name_matches_list_of_wildcard_names(file_name, self.allowed_read_files) or \
            self._is_system_file(file_name)

    def is_allowed_write_file(self, file_name: str) -> bool:
        try:
            file_path = Path(file_name).resolve()
        except TypeError:
            return True
        if self.allowed_write_folder is not None and \
                file_path.parents[0].resolve() != Path(self.allowed_write_folder).resolve() and \
                not _is_temp_file(file_name):
            return False
        return self.allowed_write_files == 'all' or \
            is_name_matches_list_of_wildcard_names(file_path.name, self.allowed_write_files) or _is_temp_file(file_name)

    def open_wrapper(self, *args, **kwargs):
        file_name = args[0] if len(args) > 0 else kwargs.get('file', None)
        open_mode = args[1] if len(args) > 1 else kwargs.get('mode', 'r')
        is_opening_for_writing = open_mode in ['w', 'a', 'x', 'w+b', 'a+b', 'x+b', 'wb', 'ab', 'xb']
        # allow read/write files when importing packages
        if not ModifyImport.get_runtime_instance().is_currently_importing():
            if is_opening_for_writing:
                if not self.is_allowed_write_file(file_name):
                    raise CodeWriteForbiddenFile(file=file_name)
            else:
                if not self.is_allowed_read_file(file_name):
                    raise CodeReadForbiddenFile(file=file_name)
        return self.original_open(*args, **kwargs)

    def _is_system_file(self, file_name):
        abs_path_to_file = os.path.abspath(file_name)
        return any(abs_path_to_file.startswith(folder) for folder in self.SYSTEM_FOLDERS) or \
            any(abs_path_to_file.endswith(file) for file in self.SYSTEM_FILES)


class FileAndMetadata(NamedTuple):
    file: str
    metadata: Any


@dataclass
class TrackCreatedFiles(SingletonRegisteredRunContext):
    output_file_requirements: Optional[OutputFileRequirements] = None  # None means allow all

    created_files: Optional[ListBasedSet[str]] = None  # None - unknown, context is not yet exited
    un_allowed_created_files: Optional[List[str]] = None  # None - unknown, context is not yet exited
    _preexisting_files_and_metadata: ListBasedSet[FileAndMetadata] = None

    def _get_dir_files_and_metadata(self) -> ListBasedSet[FileAndMetadata]:
        return ListBasedSet((file, os.stat(file).st_mtime) for file in os.listdir())

    def __enter__(self):
        self._preexisting_files_and_metadata = self._get_dir_files_and_metadata()
        self.created_files = None
        self.un_allowed_created_files = None
        return super().__enter__()

    def _get_created_files(self) -> ListBasedSet[str]:
        files_and_metadata = self._get_dir_files_and_metadata() - self._preexisting_files_and_metadata
        return ListBasedSet(file for file, _ in files_and_metadata)

    def _create_issues_for_num_files(self):
        for requirement, output_files \
                in self.output_file_requirements.get_requirements_to_output_files(self.created_files).items():
            if len(output_files) < requirement.minimal_count:
                # The specified number of output files were not created.
                if requirement.is_wildcard():
                    issue = dedent_triple_quote_str(f"""
                        The code was supposed to create at least {requirement.minimal_count} files \t
                        of "{requirement.generic_filename}", \t
                        but it only created {len(output_files)} files of this type.
                        """)
                else:
                    issue = f"The code didn't generate the desired output file, '{requirement.generic_filename}'."
                self.issues.append(RunIssue(
                    category='Not all required files were created',
                    issue=issue,
                    code_problem=CodeProblem.MissingOutputFiles,
                    comment='Code did not create all required files'
                ))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.created_files = self._get_created_files()
        if self.output_file_requirements is not None:
            self._create_issues_for_num_files()
            self.un_allowed_created_files = self.output_file_requirements.get_unmatched_files(self.created_files)
        else:
            self.un_allowed_created_files = []
        if self.un_allowed_created_files:
            raise UnAllowedFilesCreated(un_allowed_files=list(self.un_allowed_created_files))

        return super().__exit__(exc_type, exc_val, exc_tb)


@dataclass
class ModifyImport(SingletonRegisteredRunContext):
    modified_imports: Iterable[Tuple[str, Optional[str]]] = field(default_factory=list)

    _currently_importing: list = field(default_factory=list)
    original_import: Optional[Callable] = None

    def __enter__(self):
        self._currently_importing = []
        return super().__enter__()

    def _reversible_enter(self):
        self.original_import = builtins.__import__
        builtins.__import__ = self.custom_import
        super()._reversible_enter()

    def _reversible_exit(self):
        builtins.__import__ = self.original_import
        self.original_import = None
        super()._reversible_exit()

    def is_currently_importing(self) -> bool:
        return len(self._currently_importing) > 0

    def custom_import(self, name, globals=None, locals=None, fromlist=(), level=0):
        if self._is_called_from_user_script():
            matched_module, new_name = \
                next(((module, new_name) for module, new_name in self.modified_imports
                      if name.startswith(module)), (None, None))
            if matched_module:
                if new_name is None:
                    raise CodeImportForbiddenModule(module=name)
                name = new_name + name[len(matched_module):]
        with self.within_import(name):
            try:
                return self.original_import(name, globals, locals, fromlist, level)
            except Exception as e:
                e.fromlist = fromlist
                raise e

    @contextmanager
    def within_import(self, package_name):
        """
        Context manager for tracking when we enter an import statement.
        """
        self._currently_importing.append(package_name)
        try:
            yield
        finally:
            self._currently_importing.pop()


@dataclass
class WarningHandler(SingletonRegisteredRunContext):
    categories_to_issue: Optional[Iterable[Type[Warning]]] = ()
    categories_to_raise: Optional[Iterable[Type[Warning]]] = ()
    categories_to_ignore: Optional[Iterable[Type[Warning]]] = ()
    original_showwarning: Optional[Callable] = None

    def _reversible_enter(self):
        self.original_showwarning = warnings.showwarning
        warnings.showwarning = self._warning_handler
        super()._reversible_enter()

    def _reversible_exit(self):
        warnings.showwarning = self.original_showwarning
        self.original_showwarning = None
        super()._reversible_exit()

    @staticmethod
    def _is_matched_cls(category, classes):
        return classes is ... or any(issubclass(category, cls) for cls in classes)

    def _warning_handler(self, message, category, filename, lineno, file=None, line=None):
        if self._is_matched_cls(category, self.categories_to_raise):
            raise message
        elif self._is_matched_cls(category, self.categories_to_ignore):
            pass
        elif self._is_matched_cls(category, self.categories_to_issue):
            self.issues.append(RunIssue.from_current_tb(
                category='Undesired warning',
                issue=f'Code produced an undesired warning:\n```\n{str(message).strip()}\n```',
                instructions='Please see if you understand the cause of this warning and fix the code.\n'
                             'Alternatively, if the warning is expected, then change the code to ignore it.',
                code_problem=CodeProblem.NonBreakingRuntimeIssue,
            ))
        else:
            return self.original_showwarning(message, category, filename, lineno, file, line)


@dataclass
class RunInDirectory(SingletonRegisteredRunContext):
    """
    Run code in a specific folder.
    If folder is None, run in the current folder.
    """
    folder: Union[Path, str] = None
    original_cwd: Optional[str] = None

    def __enter__(self):
        self.original_cwd = os.getcwd()
        if self.folder is not None:
            os.chdir(self.folder)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.original_cwd)
        self.original_cwd = None
        return super().__exit__(exc_type, exc_val, exc_tb)



from typing import Optional


class Symbols:
    CHECK_SYMBOL = '✅'
    CROSS_SYMBOL = '❌'
    WARNING_SYMBOL = '⚠️'
    INFO_SYMBOL = 'ℹ️'

    @staticmethod
    def get_is_ok_symbol(is_ok: Optional[bool]) -> str:
        if is_ok is None:
            return Symbols.WARNING_SYMBOL
        return Symbols.CHECK_SYMBOL if is_ok else Symbols.CROSS_SYMBOL



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, List, Set

from data_to_paper.text.highlighted_text import red_text
from data_to_paper.base_cast import Agent

from .actions_and_conversations import Action
from .message import Message
from .conversation import Conversation
from .message_designation import GeneralMessageDesignation, SingleMessageDesignation, \
    convert_general_message_designation_to_int_list
from .actions_and_conversations import Conversations

NoneType = type(None)


@dataclass(frozen=True)
class ConversationAction(Action):
    """
    Base class for actions performed on a conversation.
    """

    conversations: Conversations = None

    conversation_name: Optional[str] = None
    "The name of the conversation to perform the action on."

    driver: Optional[str] = None
    "The algorithm performing the action."

    comment: Optional[str] = None
    "A comment explaining why action is performed."

    @property
    def conversation(self) -> Conversation:
        return self.conversations.get_conversation(self.conversation_name)

    def _pretty_attrs(self) -> str:
        return ''

    def pretty_repr(self, is_color: bool = True, with_conversation_name: bool = True) -> str:
        s = ''
        if self.driver:
            s += f'{self.driver}: '
        s += super().pretty_repr(is_color=is_color)
        if with_conversation_name and self.conversation_name:
            s += f' -> {self.conversation_name}'
        if self.comment:
            s += f', {self.comment}'
        if is_color:
            s = red_text(s)
        return s


@dataclass(frozen=True)
class ChangeConversationParticipants(ConversationAction):
    """
    Create a new conversation.
    """
    participants: Set[Agent] = None

    def _pretty_attrs(self) -> str:
        return f'name="{self.conversation_name}", ' \
               f'participants={sorted([p.name if isinstance(p, Agent) else p for p in self.participants])}'


class CreateConversation(ChangeConversationParticipants):
    """
    Create a new conversation.
    """

    def apply(self):
        if self.conversation_name is None:
            return
        self.conversations.get_or_create_conversation(
            conversation_name=self.conversation_name, participants=self.participants)


class AddParticipantsToConversation(ChangeConversationParticipants):
    """
    Add participants to a conversation.
    """

    def apply(self):
        for participant in self.participants:
            self.conversation.add_participant(participant)


@dataclass(frozen=True)
class ChangeMessagesConversationAction(ConversationAction):
    pass


@dataclass(frozen=True)
class AppendMessage(ChangeMessagesConversationAction):
    """
    Append a message to the conversation.

    Message will be tagged with `tag`.
    If `tag` already exists, conversation will reset to the previous tag.
    """

    message: Message = None

    delay: float = None

    def _get_index_of_tag(self) -> Optional[int]:
        """
        Return the index of the message with the provided tag.
        """
        if self.message.tag is None:
            return None
        try:
            return SingleMessageDesignation(self.message.tag).get_message_num(self.conversation)
        except ValueError:
            return None

    def _get_message_index(self):
        """
        Return the index of the message that will be appended to the conversation.
        """
        tag_index = self._get_index_of_tag()
        if tag_index is None:
            return len(self.conversation)
        return tag_index

    def should_add_to_conversation(self) -> bool:
        """
        Return True if the message should be added to the conversation.
        """
        return self.conversation_name is not None and not self.message.ignore

    def pretty_repr(self, is_color: bool = True, with_conversation_name: bool = True,
                    abbreviate_content: bool = False) -> str:
        # Note 1: the conversation len assumes this method is called right before the message is appended.
        # Note 2: we are only adding the text from the super method we have comments or are rewinding. Otherwise, we
        #         the message we print has the other information (conversation name and role).
        if not self.should_add_to_conversation():
            return ''
        s = ''
        if self.comment or self._pretty_attrs():
            s += super().pretty_repr(is_color=is_color, with_conversation_name=False) + '\n'
        s += self.message.pretty_repr(number=self._get_message_index() + 1,
                                      conversation_name=self.conversation_name,
                                      is_color=is_color, abbreviate_content=abbreviate_content)
        return s

    def _pretty_attrs(self) -> str:
        index = self._get_index_of_tag()
        if index is None:
            return ''
        num_deleted = len(self.conversation) - index
        return f'REWINDING {num_deleted} MESSAGES'

    def apply(self):
        """
        Append a message to the conversation.
        Reset the conversation to the previous tag if the tag already exists.
        """
        if not self.should_add_to_conversation():
            return
        message_index = self._get_message_index()
        index = self._get_index_of_tag()
        if index is not None:
            del self.conversation[index:]
        assert len(self.conversation) == message_index
        self.message.index_in_conversation = len(self.conversation)
        self.message.effective_index_in_conversation = len(self.conversation.get_chosen_indices_and_messages())
        self.conversation.append(self.message)


@dataclass(frozen=True)
class BaseLLMResponse(ChangeMessagesConversationAction):
    """
    Base class for an action involving getting a response from LLM.
    """

    hidden_messages: GeneralMessageDesignation = None
    "list of message to remove from the conversation when sending to the LLM"

    def _pretty_attrs(self) -> str:
        return f'HIDING MESSAGES: {self.hidden_messages}' if self.hidden_messages else ''


@dataclass(frozen=True)
class AppendLLMResponse(AppendMessage, BaseLLMResponse):
    """
    Add a response from the LLM.
    """
    pass


@dataclass(frozen=True)
class FailedLLMResponse(BaseLLMResponse):
    """
    Failed getting a response from the LLM. Nothing is appended to the conversation.
    """

    exception: Exception = None

    def pretty_repr(self, is_color: bool = True, with_conversation_name: bool = True) -> str:
        s = f': FAILED:\n{self.exception}'
        if is_color:
            s = red_text(s)
        s = super().pretty_repr(is_color=is_color, with_conversation_name=False) + s
        return s

    def apply(self):
        pass


class NullConversationAction(ConversationAction):
    """
    Add a comment to the action list.

    The conversation is not affected by comments.
    """
    pass


@dataclass(frozen=True)
class ResetToTag(ChangeMessagesConversationAction):
    """
    Reset the conversation back to right after the specified tag.

    By default, off_set is 0, which means the message with the specified tag will not be deleted.
    """
    tag: Optional[str] = None
    off_set: int = 0

    @property
    def _single_message_designation(self) -> SingleMessageDesignation:
        return SingleMessageDesignation(tag=self.tag, off_set=self.off_set)

    @property
    def _message_num(self) -> int:
        return self._single_message_designation.get_message_num(self.conversation)

    def _pretty_attrs(self) -> str:
        num_messages_deleted = len(self.conversation) - self._message_num - 1
        return f'{self._single_message_designation} [REWINDING {num_messages_deleted} MESSAGES]'

    def apply(self):
        del self.conversation[self._message_num + 1:]


@dataclass(frozen=True)
class DeleteMessages(ChangeMessagesConversationAction):
    """
    Delete all messages between `start` and `end`

    start/end can be int (message number), or str (message tag).
    """
    message_designation: GeneralMessageDesignation = None

    def _get_indices_to_delete(self) -> List[int]:
        """
        Return the indices of the messages to delete in reverse order (to allow consistent popping).
        """
        return convert_general_message_designation_to_int_list(self.message_designation, self.conversation)[-1::-1]

    def _pretty_attrs(self) -> str:
        return f'{self.message_designation} [{len(self._get_indices_to_delete())} MESSAGES]'

    def apply(self):
        for index in self._get_indices_to_delete():
            self.conversation.pop(index)


@dataclass(frozen=True)
class ReplaceLastMessage(AppendMessage):
    """
    Replace the last message with a new message.
    """
    message: Message = None

    def _pretty_attrs(self) -> str:
        return ''

    def apply(self):
        assert self.conversation[-1].role == self.message.role
        self.conversation.pop()
        super().apply()



from dataclasses import dataclass
from pathlib import Path

from .base_products_conversers import ProductsHandler


@dataclass
class BaseFileProducer(ProductsHandler):

    output_filename: str = None

    @property
    def output_file_path(self):
        return Path(self.output_directory) / self.output_filename

    @property
    def output_file_stem(self):
        return Path(self.output_file_path).stem



from cmath import isclose
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import List, Tuple, Optional

from data_to_paper.base_products import Products
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.nice_list import NiceList
from data_to_paper.utils.replacer import Replacer, StrOrReplacer
from data_to_paper.utils.types import ListBasedSet
from data_to_paper.utils.check_numeric_values import find_non_matching_numeric_values, is_number_legit
from data_to_paper.latex.latex_to_pdf import evaluate_latex_num_command
from data_to_paper.conversation import Message, GeneralMessageDesignation
from data_to_paper.servers.model_engine import ModelEngine
from data_to_paper.code_and_output_files.ref_numeric_values import find_hyperlinks, find_numeric_values, \
    find_matching_reference, replace_hyperlinks_with_values, TARGET, LINK

from .copier import Copier
from .result_converser import ResultConverser, Rewind, BumpModel
from .dual_converser import ReviewDialogDualConverserGPT


@dataclass
class ProductsHandler(Copier):
    """
    Base class for steps that deal with Products and may also create output files.
    """
    COPY_ATTRIBUTES = {'products', 'output_directory'}

    products: Products = None

    output_directory: Path = None  # if not None, save any output files to this directory

    def __post_init__(self):
        if self.output_directory:
            self.output_directory = Path(self.output_directory).absolute()

    def _move_files_to_output_directory(self, file_paths: List[Path]):
        """
        Move any output files to the output directory.
        """
        if self.output_directory:
            for file_path in file_paths:
                file_path.rename(self.output_directory / file_path.name)


@dataclass
class ProductsConverser(ProductsHandler, ResultConverser):

    def __post_init__(self):
        ProductsHandler.__post_init__(self)
        ResultConverser.__post_init__(self)


@dataclass
class BackgroundProductsConverser(ProductsConverser):
    """
    Base class for conversers that deal with Products.
    Allows for the addition of background information about prior products to the conversation.
    """
    background_product_fields: Tuple[str, ...] = None
    # tuple of product fields to provide background information about.
    # If empty tuple, do not provide background information.
    # if None, this instance was called into an already running conversation by another converser
    # and should not add any new background information.
    background_product_fields_to_hide: Tuple[str, ...] = None

    product_acknowledgement: str = "Thank you for the {}. \n"
    goal_noun: str = None
    goal_verb: str = None
    post_background_comment: str = 'Background messages completed. Requesting "{goal_noun}".'

    @property
    def replacer_kwargs(self):
        if self.background_product_fields is None:
            return {}
        fields_to_hide = self.background_product_fields_to_hide or ()
        return {field_: self.products.get_name(field_) for field_ in self.background_product_fields
                if field_ not in fields_to_hide and self.products.is_product_available(field_)}

    @property
    def actual_background_product_fields(self) -> Optional[Tuple[str, ...]]:
        if self.background_product_fields is None:
            return None
        return self._get_available_background_product_fields(self.background_product_fields)

    def _get_available_background_product_fields(self, product_fields: Tuple[str, ...]) -> Tuple[str, ...]:
        return tuple(product_field for product_field in product_fields
                     if self.products.is_product_available(product_field))

    @property
    def background_product_names(self) -> NiceList[str]:
        return NiceList((self.products.get_name(product_field)
                         for product_field in self.actual_background_product_fields),
                        wrap_with='"', separator=', ', last_separator=' and ')

    @property
    def actual_background_product_names(self) -> NiceList:
        if not self.actual_background_product_fields:
            return NiceList()
        return NiceList(
            [self.products.get_name(product_field) for product_field in self.actual_background_product_fields],
            wrap_with='"', separator=', ', empty_str='')

    @property
    def vertical_actual_background_product_names(self) -> NiceList:
        return NiceList([name for name in self.actual_background_product_names],
                        wrap_with='', separator='\n', empty_str='NO BACKGROUND PRODUCTS')

    def _get_product_tags(self, product_field: str) -> Tuple[str, str]:
        return f'background_{product_field}', f'background_thanks_{product_field}',

    def _get_acknowledgement_and_tag(self, product_field: str) -> Tuple[str, str]:
        thank_you_message = self.product_acknowledgement.format(self.products.get_name(product_field))
        _, tag = self._get_product_tags(product_field)
        return thank_you_message, tag

    def get_product_description_and_tag(self, product_field: str) -> Tuple[str, str]:
        product_description = self.products.get_description_for_llm(product_field)
        tag, _ = self._get_product_tags(product_field)
        return product_description, tag

    def _add_acknowledgement(self, product_field: str, is_last: bool = False):
        acknowledgement, tag = self._get_acknowledgement_and_tag(product_field)
        self.apply_append_surrogate_message(acknowledgement, tag=tag, is_background=True)

    def _add_product_description(self, product_field: str):
        product_description, tag = self.get_product_description_and_tag(product_field)
        self.apply_append_user_message(product_description, tag=tag, is_background=True)

    def _pre_populate_background(self):
        """
        Add background information to the conversation.
        """
        previous_product_items = self.actual_background_product_fields
        if previous_product_items is not None:
            assert len(self.conversation.get_chosen_messages()) == 1
            for i, product_field in enumerate(previous_product_items or []):
                is_last = i == len(previous_product_items) - 1
                self._add_product_description(product_field)
                self._add_acknowledgement(product_field, is_last=is_last)
            if self.post_background_comment:
                self.comment(self.post_background_comment, tag='after_background')
        return super()._pre_populate_background()

    def apply_get_and_append_assistant_message(self, tag: Optional[StrOrReplacer] = None,
                                               comment: Optional[StrOrReplacer] = None,
                                               is_code: bool = False, previous_code: Optional[str] = None,
                                               model_engine: Optional[ModelEngine] = None,
                                               hidden_messages: GeneralMessageDesignation = None,
                                               expected_tokens_in_response: int = None,
                                               background_product_fields_to_hide: Optional[Tuple[str, ...]] = None,
                                               **kwargs) -> Message:
        """
        Apply get and append assistant message.
        Allows hiding specific background product fields.
        """
        background_product_fields_to_hide = background_product_fields_to_hide or self.background_product_fields_to_hide
        if hidden_messages is None and background_product_fields_to_hide:
            hidden_messages = []
            for product_field in background_product_fields_to_hide:
                product_tag, thanks_tag = self._get_product_tags(product_field)
                hidden_messages.append(product_tag)
                hidden_messages.append(thanks_tag)

        return super().apply_get_and_append_assistant_message(
            tag=tag, comment=comment, is_code=is_code, previous_code=previous_code,
            model_engine=model_engine, hidden_messages=hidden_messages,
            expected_tokens_in_response=expected_tokens_in_response, **kwargs)


@dataclass
class ReviewBackgroundProductsConverser(BackgroundProductsConverser, ReviewDialogDualConverserGPT):
    """
    Base class for conversers that specify prior products and then set a goal for the new product
    to be suggested and reviewed.
    """
    suppress_printing_other_conversation: bool = False
    max_reviewing_rounds: int = 0
    termination_phrase: str = "The {goal_noun} does not require any changes"
    other_mission_prompt: Optional[str] = None  # None: use the mission_prompt
    sentence_to_add_at_the_end_of_performer_response: str = \
        'Please provide constructive feedback, or, if you are satisfied, respond with "{termination_phrase}".'

    def __post_init__(self):
        BackgroundProductsConverser.__post_init__(self)
        ReviewDialogDualConverserGPT.__post_init__(self)

    def _pre_populate_other_background(self):
        previous_product_items = self.actual_background_product_fields
        if previous_product_items is not None:
            assert len(self.other_conversation) == 1
            for i, product_field in enumerate(previous_product_items or []):
                is_last = i == len(previous_product_items) - 1
                self._add_other_product_description(product_field)
                self._add_other_acknowledgement(product_field, is_last=is_last)
        return super()._pre_populate_other_background()

    def _add_other_acknowledgement(self, product_field: str, is_last: bool = False):
        acknowledgement, tag = self._get_acknowledgement_and_tag(product_field)
        other_mission_prompt = self.other_mission_prompt or self.mission_prompt
        acknowledgement += f'\n{other_mission_prompt}' if is_last else ''
        self.apply_to_other_append_surrogate_message(acknowledgement, tag=tag, is_background=True)

    def _add_other_product_description(self, product_field: str):
        product_description, tag = self.get_product_description_and_tag(product_field)
        self.apply_to_other_append_user_message(product_description, tag=tag, is_background=True)


class CheckExtractionReviewBackgroundProductsConverser(ReviewBackgroundProductsConverser):
    product_fields_from_which_response_is_extracted: Tuple[str, ...] = None
    only_warn_about_non_matching_values: bool = False

    _number_of_non_matching_values: int = None

    warning_about_non_matching_values: str = dedent_triple_quote_str("""
        ########################
        ####### WARNING: #######
        ########################
        Some of the specified values {} are not explicitly extracted from:
        {names_of_products_from_which_to_extract}
        """)

    report_non_match_prompt: str = dedent_triple_quote_str("""
        Any numeric value in your section must be based on the `provided data` above, namely on numerical values \t
        extracted from: 
        {names_of_products_from_which_to_extract}

        However, upon reviewing your section, I've identified certain `potentially problematic values`, \t
        which don't directly match the `provided data`. They are: 
        {}

        For transparency, please revise your section such that it includes only values \t
        explicitly extracted from the `provided data` above, or derived from them using the \t
        `\\num{<formula>, "explanation"}` syntax. 

        Examples:
        - If you would like to report the difference between two provided values 87 and 65, you should write:
        "The initial price of 87 was changed to 65, representing a difference of \\num{87 - 65}"

        - If you would like to report the odds ratio corresponding to a provided regression coefficient of 1.234, \t
        you should write:
        "The regression coefficient was 1.234 corresponding to an odds ratio of \\num{exp(1.234)}"

        - If the provided data includes a distance of 9.1e3 cm, and you would like to report the distance in meters, \t
        you should write: 
        "Our analysis revealed a distance of \\num{9.1e3 / 100} meters"

        IMPORTANT NOTE:
        If we need to include a numeric value that was not calculated or is not explicitly given in the \t
        Display Items or "{additional_results}", \t
        and cannot be derived from them, \t
        then indicate `[unknown]` instead of the numeric value. 

        For example:

        "The regression coefficient for the anti-cancer drugs was [unknown]."
        """)

    def _get_text_from_which_response_should_be_extracted(self) -> str:
        return '\n'.join(self.products.get_description(product_field)
                         for product_field in self.product_fields_from_which_response_is_extracted
                         if self.products.is_product_available(product_field))

    @property
    def names_of_products_from_which_to_extract(self) -> List[str]:
        return NiceList((self.products.get_name(product_field)
                        for product_field in self.product_fields_from_which_response_is_extracted),
                        wrap_with='"',
                        last_separator=' and ')

    def _check_extracted_numbers(self, text: str,
                                 ignore_int_below: int = 100,
                                 remove_trailing_zeros: bool = True,
                                 allow_truncating: bool = True):
        if self.product_fields_from_which_response_is_extracted is None:
            return

        # Find the non-matching values:
        non_matching, matching = find_non_matching_numeric_values(
            source=self._get_text_from_which_response_should_be_extracted(),
            target=text,
            ignore_int_below=ignore_int_below,
            remove_trailing_zeros=remove_trailing_zeros,
            ignore_one_with_zeros=True, ignore_after_smaller_than_sign=True,
            allow_truncating=allow_truncating)
        number_of_non_matching_values, number_of_matching_values = len(non_matching), len(matching)
        if self._number_of_non_matching_values is None:
            add_iterations = 3  # first time, we start with added 3 iterations
        else:
            add_iterations = int(number_of_non_matching_values < self._number_of_non_matching_values)

        # Print to the console the number of non-matching values:
        self.comment(f'Checking {number_of_matching_values + number_of_non_matching_values} numerical values. '
                     f'Found {number_of_non_matching_values} non-matching.', as_action=False)
        if self._number_of_non_matching_values is not None:
            self.comment(f'Compared to {self._number_of_non_matching_values} non-matching in the previous iteration '
                         f'(add_iterations: {add_iterations})', as_action=False)
        self._number_of_non_matching_values = number_of_non_matching_values
        if non_matching:
            if self.only_warn_about_non_matching_values:
                self.comment(Replacer(self, self.warning_about_non_matching_values, args=(non_matching,)),
                             as_action=False)
            else:
                self._raise_self_response_error(
                    title='# Non-matching numeric values in the section',
                    error_message=Replacer(self, self.report_non_match_prompt, args=(ListBasedSet(non_matching),)),
                    rewind=Rewind.AS_FRESH,
                    add_iterations=add_iterations,
                    bump_model=BumpModel.HIGHER_STRENGTH,
                )

    def _check_url_in_text(self, text: str) -> str:
        """
        Check that the text does not contain a URL. If it does raise an error.
        """
        if 'http' in text or 'www.' in text or 'mailto' in text:
            self._raise_self_response_error(
                title='# URL in text',
                error_message='The text contains URLs which is not allowed.',
                rewind=Rewind.AS_FRESH,
            )
        return text


@dataclass
class CheckReferencedNumericReviewBackgroundProductsConverser(CheckExtractionReviewBackgroundProductsConverser):
    should_apply_numeric_referencing_to_other: bool = False
    self_products_to_other_products: Tuple[Tuple[str, str]] = ()

    report_non_match_prompt: str = dedent_triple_quote_str("""
        Your section contains some improperly referenced numeric values, specifically:

        {}

        Numeric values must be included with \\hyperlink matching the \\hypertarget in the provided sources above.
        The hyperlinks must include only the numeric values.
        For example: 
        - Correct syntax: 'P $<$ \\hyperlink{Z3c}{1e-6}'
        - Incorrect syntax: 'P \\hyperlink{Z3c}{$<$ 1e-6}'

        See the examples I provided in my previous message. 

        Remember, you can also include such hyperlinked numeric values within the <formula> of \t
        \\num{<formula>, "explanation"}.
        This allows you to derive new numeric values from the provided source data.
        Changing units, calculating differences, converting regression coefficients to odds ratios, etc. 
        For example:
        'The treatment odds ratio was \\num{exp(\\hyperlink{Z3a}{0.17}), \t
        "Translating the treatment regression coefficient to odds ratio"}'

        In summary: 
        Either provided as a stand alone or within the <formula> of \\num{<formula>, "explanation"}, \t
        all numeric values must have \\hyperlink references \t
        that match the \\hypertarget references in the provided sources above.

        IMPORTANT NOTE:
        If we need to include a numeric value that is not explicitly provided in the Display Items and other results \t
        above, and cannot be derived from them, then indicate `[unknown]` instead of the numeric value. 

        For example:
        'The p-value of the regression coefficient of the treatment was [unknown].'
        """)

    def _get_text_from_which_response_should_be_extracted(self) -> str:
        return '\n'.join(self.products.get_description_for_llm(product_field)
                         for product_field in self.product_fields_from_which_response_is_extracted
                         if self.products.is_product_available(product_field))

    def _replace_product_field_from_self_to_other(self, product_field: str) -> str:
        if not self.should_apply_numeric_referencing_to_other:
            self_products_to_other_products = dict(self.self_products_to_other_products)
            return self_products_to_other_products.get(product_field, product_field)
        return product_field

    def _add_other_product_description(self, product_field: str):
        product_field = self._replace_product_field_from_self_to_other(product_field)
        super()._add_other_product_description(product_field)

    def _add_other_acknowledgement(self, product_field: str, is_last: bool = False):
        product_field = self._replace_product_field_from_self_to_other(product_field)
        super()._add_other_acknowledgement(product_field, is_last=is_last)

    def _alter_self_response(self, response: str) -> str:
        """
        We modify the self response so that the reviewer does not need to deal with hyperlinks or
        wit the "explanation" in \num{}.
        """
        if not self.should_apply_numeric_referencing_to_other:
            response = replace_hyperlinks_with_values(response, is_targets=False)
            response = evaluate_latex_num_command(response, just_strip_explanation=True)[0]
        response = super()._alter_self_response(response)
        return response

    def _check_extracted_numbers(self, text: str,
                                 ignore_int_below: int = 100,
                                 remove_trailing_zeros: bool = True,
                                 allow_truncating: bool = True):
        if self.product_fields_from_which_response_is_extracted is None:
            return

        if TARGET.replace('\\', '') in text:
            self._raise_self_response_error(
                title='# Wrong numeric referencing',
                error_message=f'Do not use `{TARGET}`, use `{LINK}` instead.',
                rewind=Rewind.AS_FRESH,
            )

        source_hypertargets = find_hyperlinks(self._get_text_from_which_response_should_be_extracted(), is_targets=True)

        numeric_values_without_hyperlinks = find_numeric_values(text, remove_hyperlinks=True)
        numeric_values_without_hyperlinks = [v for v in numeric_values_without_hyperlinks
                                             if not is_number_legit(v, ignore_int_below=100)]
        hyperlinked_values = find_hyperlinks(text, is_targets=False)
        hyperlinked_values_not_numeric = []
        hyperlinked_values_with_no_matching_target = []
        hyperlinked_values_not_matching_target_values = []
        for hyperlinked_value in hyperlinked_values:
            matching_target = find_matching_reference(hyperlinked_value, source_hypertargets)
            float_value = hyperlinked_value.to_float()
            if matching_target is None:
                # the reference does not exist in the provided data
                hyperlinked_values_with_no_matching_target.append(hyperlinked_value)
            elif float_value is None:
                # the value is not numeric
                hyperlinked_values_not_numeric.append(hyperlinked_value)
            elif not (isclose(abs(float_value), abs(matching_target.to_float()))
                      or isclose(abs(float_value), abs(matching_target.to_float()) * 100)):
                hyperlinked_values_not_matching_target_values.append((hyperlinked_value, matching_target))
        s = ''
        nice_list = partial(NiceList, wrap_with='"', separator='\n', last_separator=None)
        if numeric_values_without_hyperlinks:
            s += f'Some numeric values appear without a hyperlink:\n' \
                 f'{numeric_values_without_hyperlinks}\n\n'
        if hyperlinked_values_not_numeric:
            s += f'Some hyperlinks have values that are not purely numeric:\n' \
                 f'{nice_list(hyperlinked_values_not_numeric)}\n\n'
        if hyperlinked_values_with_no_matching_target:
            s += f'Some hyperlinks have labels that do not exist as hypertarget in our `provided data`:\n' \
                 f'{nice_list(hyperlinked_values_with_no_matching_target)}\n\n'
        if hyperlinked_values_not_matching_target_values:
            s += f'Some hyperlinks have values that do not exactly match the hypertarget values:\n'
            for hyperlink, hypertarget in hyperlinked_values_not_matching_target_values:
                s += f'"{hyperlink}" not matching "{hypertarget}"\n'
            s += '\n'

        if s:
            self._raise_self_response_error(
                title='# Wrong numeric referencing',
                error_message=Replacer(self, self.report_non_match_prompt, args=(s,)),
                rewind=Rewind.AS_FRESH,
                bump_model=BumpModel.HIGHER_STRENGTH,
            )



import re
from dataclasses import dataclass, field

from typing import Optional, List, Tuple

from data_to_paper.env import WRITING_MODEL_ENGINE

from data_to_paper.text import dedent_triple_quote_str, wrap_as_block
from data_to_paper.utils.nice_list import NiceList
from data_to_paper.latex import (
    FailedToExtractLatexContent,
    extract_latex_section_from_response,
)
from data_to_paper.latex.exceptions import (
    UnwantedCommandsUsedInLatex,
    BaseLatexProblemInCompilation,
)
from data_to_paper.run_gpt_code.code_utils import (
    extract_content_of_triple_quote_block,
    FailedExtractingBlock,
    IncompleteBlockFailedExtractingBlock,
    NoBlocksFailedExtractingBlock,
)
from data_to_paper.latex.citataion_utils import find_citation_ids
from data_to_paper.utils.types import ListBasedSet
from data_to_paper.latex.latex_doc import LatexDocument
from data_to_paper.latex.clean_latex import (
    process_latex_text_and_math,
    check_usage_of_un_allowed_commands,
)
from data_to_paper.latex.latex_section_tags import (
    get_list_of_tag_pairs_for_section_or_fragment,
    SECTIONS_OR_FRAGMENTS_TO_TAG_PAIR_OPTIONS,
)
from data_to_paper.servers.model_engine import ModelEngine

from .base_products_conversers import ReviewBackgroundProductsConverser
from .result_converser import Rewind


def is_similar_bibtex_ids(incorrect_id: str, correct_id: str) -> bool:
    """
    Return whether the two bibtex ids are similar.
    """
    correct_id = correct_id.lower()
    incorrect_id = incorrect_id.lower()
    return correct_id.startswith(incorrect_id) or incorrect_id.startswith(correct_id)


def replace_word(text, original_word, new_word):
    pattern = r"\b" + re.escape(original_word) + r"\b"
    return re.sub(pattern, new_word, text)


def remove_citations_from_section(section: str) -> str:
    r"""
    Return the section without citations (e.g. \cite{...}).
    """
    return re.sub(
        pattern=r"\s*\\cite[tp]?(\[.*?])?(\[.*?])?\{[^}]*}(?=\s*\.)?",
        repl="",
        string=section,
    )


@dataclass
class LatexReviewBackgroundProductsConverser(ReviewBackgroundProductsConverser):
    """
    A base class for agents requesting the LLM to write one or more latex sections.
    Option for removing citations from the section.
    Option for reviewing the sections (set max_review_turns > 0).
    """

    latex_document: LatexDocument = field(default_factory=LatexDocument)
    model_engine: ModelEngine = WRITING_MODEL_ENGINE
    should_remove_citations_from_section: bool = True

    section_names: List[str] = field(default_factory=list)

    your_response_should_be_formatted_as: str = "a triple backtick latex block."
    formatting_instructions_for_feedback: str = dedent_triple_quote_str(
        """
        Please {goal_verb} the {goal_noun} again according to my feedback above.

        Remember, your response should be formatted as {your_response_should_be_formatted_as}
        """
    )
    rewind_after_getting_a_valid_response: Optional[Rewind] = Rewind.AS_FRESH

    un_allowed_commands: Tuple[str, ...] = (r"\cite", r"\verb", r"\begin{figure}")

    response_to_non_matching_citations: str = dedent_triple_quote_str(
        """
        The following citation ids were not found: 
        {}
        Please make sure all citation ids are writen exactly as in the citation lists above.
        """
    )

    response_to_floating_citations: str = dedent_triple_quote_str(
        """
        The following citation ids are not properly enclosed in \\cite{{}} command: 
        {}
        Please make sure all citation ids are enclosed in a \\cite{{}} command.
        """
    )

    @property
    def section_name(self) -> Optional[str]:
        if len(self.section_names) == 1:
            return self.section_names[0]
        return None

    @property
    def pretty_section_names(self) -> NiceList[str]:
        """
        Return the section names capitalized.
        """
        return NiceList(
            (section_name.title() for section_name in self.section_names),
            separator=", ",
            last_separator=" and ",
            wrap_with="`",
        )

    def get_valid_result_as_markdown(self) -> str:
        return wrap_as_block("\n\n".join(self.valid_result), "latex")

    def _convert_extracted_text_to_fresh_looking_response(
        self, extracted_text: List[str]
    ) -> str:
        """
        Return a response that looks fresh.
        """
        s = super()._convert_extracted_text_to_fresh_looking_response(extracted_text)
        s = wrap_as_block(s, "latex")
        return s

    def _convert_valid_result_back_to_extracted_text(
        self, valid_result: List[str]
    ) -> List[str]:
        return valid_result

    def _get_allowed_bibtex_citation_ids(self) -> List[str]:
        r"""
        Return the bibtex citation ids that are allowed citing with \cite{}.
        """
        return []

    def _extract_latex_section_from_response(
        self, response: str, section_name: str
    ) -> str:
        """
        Extract a section from the response.
        """
        tags_list = get_list_of_tag_pairs_for_section_or_fragment(section_name)
        tags = tags_list[0]
        is_flanking_tags = len(tags_list) == 1 and tags.is_flanking()
        try:
            response = extract_content_of_triple_quote_block(response, "latex", "latex")
        except FailedExtractingBlock as e:
            if isinstance(e, NoBlocksFailedExtractingBlock) and is_flanking_tags:
                pass
            else:
                self._raise_self_response_error(
                    title="# Failed to extract latex block",
                    error_message=str(e),
                    missing_end=isinstance(e, IncompleteBlockFailedExtractingBlock),
                )
        try:
            return extract_latex_section_from_response(response, section_name)
        except FailedToExtractLatexContent as e:
            self._raise_self_response_error(
                title="# Failed to extract latex section(s)",
                error_message=str(e),
                missing_end=is_flanking_tags
                and tags[0] in response
                and tags[1] not in response,
            )

    def _process_non_math_parts(self, section: str) -> str:
        return process_latex_text_and_math(section)

    def _check_usage_of_un_allowed_commands(self, section: str) -> str:
        try:
            check_usage_of_un_allowed_commands(section, self.un_allowed_commands)
        except UnwantedCommandsUsedInLatex as e:
            self._raise_self_response_error(
                title="# Unwanted commands used in latex", error_message=str(e)
            )
        return section

    def _remove_citations_from_section(self, section) -> str:
        """
        Remove the citations that the LLM inserted by mistake.
        """
        if not self.should_remove_citations_from_section:
            return section
        section = remove_citations_from_section(section)
        # also remove \bibliographystyle{} and \bibliography{} commands
        section = re.sub(
            pattern=r"\s*\\bibliographystyle\{.*?\}", repl="", string=section
        )
        section = re.sub(pattern=r"\s*\\bibliography\{.*?\}", repl="", string=section)
        return section

    def _check_and_correct_citations(self, section: str) -> str:
        """
        Check that all \\cite{} commands refer to bibtex citations that are allowed.
        Raise if citing un-allowed citations, or correct similar citations to the correct one.
        """
        allowed_ids = self._get_allowed_bibtex_citation_ids()
        not_found_ids = ListBasedSet(
            [
                citation_id
                for citation_id in find_citation_ids(section)
                if citation_id not in allowed_ids
            ]
        )
        # Correct almost-matching citations:
        not_found_and_not_corrected_ids = []
        for not_found_id in not_found_ids:
            similar_ids = [
                allowed_id
                for allowed_id in allowed_ids
                if is_similar_bibtex_ids(not_found_id, allowed_id)
            ]
            if len(similar_ids) == 1:
                section = replace_word(section, not_found_id, similar_ids[0])
                self.comment(
                    f"Replacing citation id {not_found_id} with {similar_ids[0]}",
                    as_action=False,
                )
            else:
                not_found_and_not_corrected_ids.append(not_found_id)
        if not_found_and_not_corrected_ids:
            self._raise_self_response_error(
                title="# Citation ids not found",
                error_message=self.response_to_non_matching_citations.format(
                    not_found_and_not_corrected_ids
                ),
            )
        return section

    def _check_for_floating_citations(self, section: str) -> str:
        r"""
        Check that there are no floating citations, namely citations not enclosed with \cite{}.
        """
        if self.response_to_floating_citations:
            section_without_citations = remove_citations_from_section(section)
            allowed_ids = self._get_allowed_bibtex_citation_ids()
            floating_ids = ListBasedSet(
                [
                    citation_id
                    for citation_id in allowed_ids
                    if citation_id in section_without_citations
                ]
            )
            if floating_ids:
                self._raise_self_response_error(
                    title="# Floating citation ids",
                    error_message=self.response_to_floating_citations.format(
                        floating_ids
                    ),
                )
        return section

    def _process_citations(self, section: str) -> str:
        section = self._remove_citations_from_section(section)
        section = self._check_and_correct_citations(section)
        section = self._check_for_floating_citations(section)
        return section

    def _check_and_refine_section(self, section: str, section_name: str) -> str:
        section = self._check_usage_of_un_allowed_commands(section)
        section = self._process_citations(section)
        section = self._process_non_math_parts(section)
        return section

    def _check_no_additional_sections(self, response: str):
        """
        Check that there are no additional sections in the response.
        """
        required_sections = [
            section_name.title()
            for section_name in self.section_names
            if section_name not in SECTIONS_OR_FRAGMENTS_TO_TAG_PAIR_OPTIONS
        ]
        provided_sections = re.findall(
            pattern=r"\\section\*?\{([^}]*)\}", string=response
        )
        provided_but_not_required_sections = set(provided_sections) - set(
            required_sections
        )
        if provided_but_not_required_sections:
            self._raise_self_response_error(
                title="# Undesired sections in the response",
                error_message=dedent_triple_quote_str(
                    f"""
                You must only write the {self.pretty_section_names} section(s).
                But, you wrote also these section(s):
                {NiceList(provided_but_not_required_sections, wrap_with="`")}
                """
                ),
            )

        # check for duplicates in provided_sections:
        sections_appearing_more_than_once = [
            section
            for section in provided_sections
            if provided_sections.count(section) > 1
        ]
        if sections_appearing_more_than_once:
            self._raise_self_response_error(
                title="# Duplicate sections in the response",
                error_message=dedent_triple_quote_str(
                    f"""
                You wrote the following section(s) more than once:
                {NiceList(sections_appearing_more_than_once, wrap_with="`")}
                """
                ),
            )

    def _check_response_and_get_extracted_text(self, response: str) -> List[str]:
        """
        Check the response from self and extract the needed information into extracted_text.
        """
        self._check_no_additional_sections(response)
        return [
            self._extract_latex_section_from_response(response, section_name)
            for section_name in self.section_names
        ]

    def _check_extracted_text_and_update_valid_result(self, extracted_text: List[str]):

        # check and refine the sections
        for i in range(len(extracted_text)):
            extracted_text[i] = self._check_and_refine_section(
                extracted_text[i], self.section_names[i]
            )

        # check the latex compilation
        try:
            _, _, width = self.latex_document.compile_document(
                {
                    section_name: section
                    for section_name, section in zip(self.section_names, extracted_text)
                }
            )
        except BaseLatexProblemInCompilation as e:
            self._raise_self_response_error(
                title="# LaTex compilation error", error_message=str(e)
            )

        # store the result if there are no exceptions:
        self._update_valid_result(extracted_text)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, List, Tuple, Union, Type, Dict, Any, Iterable

import numpy as np

from data_to_paper.env import SUPPORTED_PACKAGES, DEBUG_MODE, MAX_EXEC_TIME, PAUSE_AT_RULE_BASED_FEEDBACK
from data_to_paper.text import dedent_triple_quote_str, line_count, wrap_as_block
from data_to_paper.utils.replacer import format_value
from data_to_paper.utils.print_to_file import print_and_log
from data_to_paper.base_cast import Agent
from data_to_paper.conversation.message_designation import RangeMessageDesignation
from data_to_paper.code_and_output_files.code_and_output import CodeAndOutput
from data_to_paper.code_and_output_files.output_file_requirements import BaseContentOutputFileRequirement, \
    OutputFileRequirements

from data_to_paper.interactive.app_interactor import _raise_if_reset

from data_to_paper.run_gpt_code.run_issues import CodeProblem, RunIssue, RunIssues
from data_to_paper.run_gpt_code.code_runner_wrapper import CodeRunnerWrapper
from data_to_paper.run_gpt_code.code_utils import FailedExtractingBlock, IncompleteBlockFailedExtractingBlock
from data_to_paper.run_gpt_code.exceptions import FailedRunningCode, UnAllowedFilesCreated, \
    CodeUsesForbiddenFunctions, CodeWriteForbiddenFile, CodeReadForbiddenFile, CodeImportForbiddenModule
from data_to_paper.interactive import PanelNames, Symbols
from data_to_paper.run_gpt_code.known_mis_imports import KNOWN_MIS_IMPORTS
from data_to_paper.run_gpt_code.base_run_contexts import RunContext
from data_to_paper.run_gpt_code.code_runner import CodeRunner
from data_to_paper.run_gpt_code.extract_and_check_code import get_issue_for_use_of_a_forbidden_function, CodeExtractor

from .base_products_conversers import BackgroundProductsConverser


def _get_description_of_run_error(error: Exception):
    str_error = str(error)
    if len(str_error) > 2000:
        str_error = str_error[:1000] + '\n[...]\n' + str_error[-800:]
    return dedent_triple_quote_str("""
        I ran the code and got the following error message:
        ```error
        {}
        ```
        """).format(str_error)


# Good for debugging:
FORGIVE_ALL_FORGIVABLE_ISSUES = False


@dataclass
class DebuggerConverser(BackgroundProductsConverser):
    """
    Interact with the LLM to debug a code that needs to create an output file.

    Starting with a conversation which ends with a code-request from the user, DebuggerConverser interacts
    with the LLM to enhance the code until it runs properly and creates a desired output file.

    Interactions with LLM include adequate reporting of:
    * missing packages
    * trying to load wrong files
    * syntax errors
    * runtime exceptions
    * too long runs (timeout)
    * output file not created
    """

    # input files:
    data_folder: Path = None
    data_filenames: Optional[list] = field(default_factory=list)

    # output files:
    output_file_requirements: OutputFileRequirements = OutputFileRequirements()

    mission_prompt: str = None
    assistant_agent: Agent = None
    user_agent: Agent = None

    supported_packages: Tuple[str, ...] = SUPPORTED_PACKAGES
    headers_required_in_code: Tuple[str, ...] = ()

    prompt_to_append_at_end_of_response: str = \
        dedent_triple_quote_str("""
            # Instructions:
            Please rewrite the complete code again with these issues corrected.

            # General formatting instructions:
            Even if you are changing just a few lines, you must return the complete code again in a \t
            single code block, including the unchanged parts, so that I can just copy-paste and run it.
            {required_headers_prompt}    
        """)

    max_debug_iterations: int = 5
    debug_iteration = 0
    timeout_sec: int = MAX_EXEC_TIME.val
    code_extractor_cls: Type[CodeExtractor] = CodeExtractor
    code_runner_cls: Type[CodeRunner] = CodeRunner
    additional_contexts: Dict[str, RunContext] = field(default_factory=dict)
    previous_code: Optional[str] = None
    _requesting_small_change: bool = False  # True when USER ask for modifications of an already existing code
    previous_code_problem: CodeProblem = CodeProblem.NoCode
    code_and_output_cls: Type[CodeAndOutput] = CodeAndOutput

    issues_to_counts: Dict[RunIssue, int] = field(default_factory=dict)
    """
    PROPERTIES
    """

    @property
    def required_headers_prompt(self) -> str:
        if len(self.headers_required_in_code) == 0:
            return ''
        return 'Remember, your code must contain the following sections:\n' + \
               '\n'.join(f'"{header}"' for header in self.headers_required_in_code)

    @property
    def iteration_str(self):
        return f'Debug iteration {self.debug_iteration}/{self.max_debug_iterations}'

    @property
    def description_of_allowed_output_files(self):
        requirements = self.output_file_requirements
        if len(requirements) == 0:
            return 'Your code should not write to any file.'

        return 'Your code should only write to these files: {}.'.format(
            ', '.join(f'"{r.generic_filename}"' for r in requirements)
        )

    def _get_allowed_packages(self, contexts) -> str:
        import_context = contexts['ModifyImport']
        return self.supported_packages + import_context.get_custom_imports()

    """
    ISSUES
    """

    def _get_issue_for_known_mis_imports(self, error: ImportError, allowed_packages) -> Optional[RunIssue]:
        if not hasattr(error, 'fromlist'):
            return
        if error.fromlist is None:
            return
        if len(error.fromlist) != 1:
            return
        var = error.fromlist[0]
        if var not in KNOWN_MIS_IMPORTS:
            return
        correct_package = KNOWN_MIS_IMPORTS[var]
        # extract from correct_package up to the first '.':
        package_base = correct_package[:correct_package.index('.')] if '.' in correct_package else correct_package
        if package_base not in self.supported_packages:
            return
        return RunIssue(
            category='Importing packages',
            issue=_get_description_of_run_error(error),
            instructions=dedent_triple_quote_str("""
                Your code should only use these packages: {allowed_packages}.
                Note that there is a `{var}` in `{known_package}`. Is this perhaps what you needed? 
                """).format(allowed_packages=allowed_packages, var=var, known_package=KNOWN_MIS_IMPORTS[var]),
            code_problem=CodeProblem.RuntimeError,
            comment='ImportError detected in gpt code',
        )

    def _get_issue_for_allowed_packages(self, error: ImportError, contexts) -> Optional[RunIssue]:
        allowed_packages = self._get_allowed_packages(contexts)
        respond_to_known_mis_imports = self._get_issue_for_known_mis_imports(error, allowed_packages)
        if respond_to_known_mis_imports:
            return respond_to_known_mis_imports
        return RunIssue(
            category='Importing packages',
            issue=_get_description_of_run_error(error),
            instructions=dedent_triple_quote_str("""
                Your code should only use these packages: {allowed_packages}.
                """).format(allowed_packages=allowed_packages),
            code_problem=CodeProblem.RuntimeError,
            comment='ImportError detected in gpt code',
        )

    def _get_issue_for_file_not_found(self, error: FileNotFoundError, e: FailedRunningCode = None) -> RunIssue:
        return RunIssue(
            category='Available input files',
            issue=_get_description_of_run_error(error),
            instructions=dedent_triple_quote_str("""
                As noted in the data description, we only have these files:
                {}  

                Note that all input files are located in the same directory as the code. 
                """).format(self.data_filenames),
            code_problem=CodeProblem.RuntimeError,
            comment='FileNotFound detected in code',
        )

    def _get_issue_for_regular_exception_or_warning(self, error: FailedRunningCode, num_added_lines: int,
                                                    ) -> RunIssue:
        return RunIssue(
            category='Runtime exception',
            issue=_get_description_of_run_error(error.get_traceback_message(num_added_lines)),
            code_problem=CodeProblem.SyntaxError if isinstance(error, SyntaxError) else CodeProblem.RuntimeError,
            comment='Runtime exception in code',
        )

    def _get_issue_for_timeout(self, error: TimeoutError, e: FailedRunningCode = None) -> RunIssue:
        linenos_lines, msg = e.get_lineno_line_message()
        on_line = '\n'.join(f'On line {lineno}: {line}' for lineno, line in linenos_lines)
        return RunIssue(
            category='Timeout',
            issue=f"I ran the code, but it just ran forever... Perhaps got stuck in too long calculations.\n"
                  f"{on_line}",
            instructions="Anything we can do to make it run faster?",
            code_problem=CodeProblem.TimeoutError,
            comment='Code has timed out',
        )

    def _get_issue_for_incomplete_code_block(self, is_bumped: bool) -> RunIssue:
        if is_bumped:
            instructions = f"Let's bump you up to {self.model_engine} and REGENERATE!"
        else:
            instructions = "Please REGENERATE!"

        return RunIssue(
            category='Code extraction problem',
            issue="Your sent incomplete code.",
            instructions=instructions,
            comment='Code is incomplete',
            end_with='',
            code_problem=CodeProblem.IncompleteBlock,
        )

    def _get_issue_for_missing_or_multiple_code_blocks(self, e: FailedExtractingBlock) -> RunIssue:
        """
        We notify missing or incomplete code to the LLM.
        If the conversation already has this notification, we regenerate gpt response instead.
        """
        return RunIssue(
            category='Code extraction problem',
            issue=str(e),
            comment='Failed extracting code from gpt response',
            end_with=self.required_headers_prompt,
            code_problem=CodeProblem.NotSingleBlock,
        )

    def _get_issue_for_forbidden_functions(self, error: CodeUsesForbiddenFunctions, e: FailedRunningCode = None
                                           ) -> RunIssue:
        func = error.func
        return get_issue_for_use_of_a_forbidden_function(func,
                                                         suggest_print_to_output=bool(self.output_file_requirements))

    def _get_issue_for_forbidden_import(self, error: CodeImportForbiddenModule, e: FailedRunningCode) -> RunIssue:
        module = error.module
        return RunIssue(
            category='Importing packages',
            issue=f"Your code import the module `{module}`, which is not allowed.",
            instructions="Your code can only use these packages: {supported_packages}.",
            code_problem=CodeProblem.RuntimeError,
            comment='Code imports forbidden module')

    def _get_issue_for_forbidden_write(self, error: CodeWriteForbiddenFile, e: FailedRunningCode) -> RunIssue:
        file = error.file
        file_and_ext = Path(file).name
        return RunIssue(
            category='Write to unallowed files',
            issue=f'Your code writes to the file "{file_and_ext}" which is not allowed.',
            instructions=self.description_of_allowed_output_files,
            code_problem=CodeProblem.RuntimeError,
            comment='Code writes to forbidden file',
        )

    def _get_issue_for_un_allowed_files_created(self, error: UnAllowedFilesCreated, e: FailedRunningCode) -> RunIssue:
        return RunIssue(
            category='Wrong output files',
            issue=f"Your code creates the following files {error.un_allowed_files} which is not allowed.",
            instructions=self.description_of_allowed_output_files,
            code_problem=CodeProblem.RuntimeError,
            comment='Code created forbidden files',
        )

    def _get_issue_for_forbidden_read(self, error: CodeReadForbiddenFile, e: FailedRunningCode) -> RunIssue:
        file = error.file
        is_read_file_in_output_file_requirements = len(self.output_file_requirements.get_unmatched_files([file])) == 0
        if is_read_file_in_output_file_requirements:
            return RunIssue(
                category='Wrong input file',
                issue=f'Your code tries reading from the output file "{file}".',
                instructions=dedent_triple_quote_str("""
                    The code can create and write to this output file, but should not read from it.
                    The only input files from which we can read the data are: 
                    {}
                    """).format(self.data_filenames),
                code_problem=CodeProblem.RuntimeError,
                comment='Code reads from output file',
            )
        else:
            return RunIssue(
                category='Wrong input file',
                issue=f'Your code reads from the file "{file}" which is not part of the dataset.',
                instructions=dedent_triple_quote_str("""
                    We only have these files:
                    {}

                    Note that all input files are located in the same directory as the code. 
                    """).format(self.data_filenames),
                code_problem=CodeProblem.RuntimeError,
                comment='Code reads from forbidden file',
            )

    def _get_issues_for_output_file_content(self, requirement: BaseContentOutputFileRequirement,
                                            filename: str, content: str) -> List[RunIssue]:
        return requirement.get_issues_for_output_file_content(filename, content)

    def _get_issues_for_created_output_files(self, code_and_output: CodeAndOutput, contexts) -> List[RunIssue]:
        issues = []
        files_to_contents = code_and_output.created_files.get_created_content_files_to_contents()
        for requirement in self.output_file_requirements:
            if isinstance(requirement, BaseContentOutputFileRequirement):
                for filename in code_and_output.created_files[requirement]:
                    issues.extend(
                        self._get_issues_for_output_file_content(requirement, filename, files_to_contents[filename]))
        return issues

    def _get_issue_for_new_code_not_being_a_modification_of_old_code(self, new_code: str,
                                                                     old_code: str) -> Optional[RunIssue]:
        if line_count(new_code) < line_count(old_code) * 0.9:
            return RunIssue(
                category='New code instead of modifications',
                issue="Your code does not seem to be a modification of the previous code.",
                instructions="Please rewrite the complete code again, making sure that the new code is "
                             "a modification of the old code.",
                comment='Code is not a modification of previous code.',
                end_with='',
                code_problem=CodeProblem.StaticCheck,
            )
        return None

    """
    METHODS FOR RUNNING CODE
    """

    def _bump_to_model_with_more_context(self) -> bool:
        try:
            self.model_engine = self.model_engine.get_model_with_more_context()
            return True
        except ValueError:
            return False

    def _get_code_runner_wrapper(self, code: str) -> CodeRunnerWrapper:
        return CodeRunnerWrapper(
            code=code,
            timeout_sec=self.timeout_sec,
            code_runner=self.get_code_runner(),
        )

    def _get_code_extractor(self) -> CodeExtractor:
        code_extractor = self.code_extractor_cls()
        if self.headers_required_in_code:
            code_extractor.headers_required_in_code = self.headers_required_in_code
        return code_extractor

    def get_code_runner(self) -> CodeRunner:
        return self.code_runner_cls(
            allowed_open_read_files=self.data_filenames,
            output_file_requirements=self.output_file_requirements,
            run_folder=self.data_folder,
            additional_contexts=self.additional_contexts,
        )

    def _get_code_and_output(self, code: str, result: str, created_files: Iterable[str],
                             contexts: Dict[str, Any] = None) -> CodeAndOutput:
        """
        Return the CodeAndOutput object for the given result and created files.
        """
        return self.code_and_output_cls(
            code=code,
            result=result,
            created_files=self.output_file_requirements.convert_to_output_file_requirements_with_content(
                created_files=created_files, run_folder=self.data_folder),
            dataframe_operations=contexts['TrackDataFrames'].dataframe_operations
            if 'TrackDataFrames' in contexts else None,
            contexts=contexts,
        )

    def _get_response_count(self) -> int:
        """
        USER: Please write code ...
        ASSISTANT: <code>   # 0
        USER: You have a bug ...
        ASSISTANT: <code>   # 1
        """
        return (len(self.conversation) - self._conversation_len_before_first_response - 1) // 2

    def _post_code_as_fresh(self, code: str, code_problem: Optional[CodeProblem] = None, action_stage: int = 0):
        self._rewind_conversation_to_first_response(offset=action_stage * 2)
        if action_stage == 0:
            self.previous_code_problem = code_problem
            message = 'Here is the code to perform the requested analysis:'
            comment = 'Code is freshly re-posted, as if it was the FIRST response.'
        elif action_stage == 1:
            message = 'Here is the revised code to perform the requested analysis:'
            comment = 'Code is freshly re-posted, as if it was the SECOND response.'
        else:
            raise ValueError(f'Invalid action_stage: {action_stage}')
        self.previous_code = code

        self.apply_append_surrogate_message(
            content=message + '\n' + wrap_as_block(code, 'python'),
            comment=comment,
        )

    def _remove_issues_exceeding_max_count(self, issues: List[RunIssue]):
        return [issue for issue in issues if issue.forgive_after is None or
                not FORGIVE_ALL_FORGIVABLE_ISSUES and
                self.issues_to_counts.get(issue, 0) < issue.forgive_after]

    def _respond_to_issues(self, issues: Union[None, RunIssue, List[RunIssue], RunIssues],
                           code_and_output: Optional[CodeAndOutput] = None,
                           is_bumped: bool = False
                           ) -> Optional[CodeAndOutput]:
        """
        We post a response to the assistant code, based on the issues.
        We also need to delete (some) of the previous exchange.

        The conversation may have a max of 3 exchanges:

        User: Please write code ... [preexisting request prompt]

        Assistant: <code>   # stage 0
        User: We have a problem ... Please fix.

        Assistant: <code>   # stage 1
        User: You have a bug ...

        Assistant: <code>   # stage 2  [We continue here only if stage 2 was a run-time error]

        In each stage, we need to decide on the action based on the problem in the code:
        - Re-post the code ["repost0" as the original response; "repost1" as the second response]
        - Leave the response as is ("leave")
        - Regenerate ("regen0", "regen1", "regen2": the original response, the second response, the third response)
        """
        if code_and_output and self.app:
            try:
                code_and_output_as_html = code_and_output.as_html()
            except Exception:  # noqa
                pass
            else:
                self._app_send_prompt(PanelNames.PRODUCT, code_and_output_as_html, provided_as_html=True)

        # Get issues
        if issues is None:
            return code_and_output
        if isinstance(issues, RunIssue):
            issues = [issues]
        if not isinstance(issues, RunIssues):
            issues = RunIssues(issues)

        # Get Problem
        problem = issues.get_most_severe_problem()

        # Get Action
        # When we have run_failed, we essentially don't know the quality of the code (e.g. it can be the
        # perfect code with just syntax error, or it can be a code doesn't even create any output files).
        # So run_failed is the only time we allow going from stage 1 to 2.
        # Namely, if we are in stage 2, we had definitely has a run_failed on stage 1.
        plan = np.array((
            # 0              1                   2      <- stage    # Problem           # noqa
            ('regen0',      'regen1',           'regen1'),          # incomplete        # noqa
            ('leave',       'regen1',           'regen2'),          # not_single_block  # noqa
            ('repost0',     'repost0/regen1',   'regen2'),          # static_check      # noqa
            ('repost0',     'repost0/leave',    'repost1'),         # run_failed        # noqa
            ('repost0',     'repost0/leave',    'repost0/regen1'),  # missing_files     # noqa
            ('repost0',     'repost0',          'repost0'),         # run_completed     # noqa
        ))
        #  xxx/yyy: xxx if current problem difficulty is equal or easier to previous_code_problem / else yyy

        current_stage = self._get_response_count()
        action = plan[problem.get_stage(), current_stage]
        if problem.get_stage() == 0 and current_stage == 2:
            assert action == 'regen1'
            if is_bumped:
                action = 'regen2'

        if '/' in action:
            action1, action2 = action.split('/')
            action = action1 if problem.get_stage() >= self.previous_code_problem.get_stage() else action2

        if DEBUG_MODE:
            print_and_log(f'=====================\n'
                          f'current_stage={current_stage}\n'
                          f'      problem={problem}\n'
                          f'prev. problem={self.previous_code_problem}\n'
                          f'       action={action}\n'
                          f'=====================\n', should_log=False)

        if action.startswith("repost") or action.startswith("regen"):
            action_stage = int(action[-1])
            action = action[:-1]
        else:
            action_stage = current_stage

        # Apply Action
        if action == "repost":
            self._post_code_as_fresh(code_and_output.code, problem, action_stage)

        message, comment, posted_issues = issues.get_message_and_comment(
            end_with=format_value(self, self.prompt_to_append_at_end_of_response))
        message += '\n\nREGENERATE' if action == "regenerate" else ''
        with self._app_temporarily_set_panel_status(PanelNames.FEEDBACK):
            self.apply_append_user_message(
                content=message,
                comment=self.iteration_str + ': ' + comment,
                sleep_for=PAUSE_AT_RULE_BASED_FEEDBACK,
            )

        if action == "regen":
            # To regenerate, we delete the required pairs of assistant+user messages
            # (including the last message which is the just-posted user response to current issue).
            posted_issues = []
            self.apply_delete_messages(
                RangeMessageDesignation.from_(start=(action_stage - current_stage - 1) * 2, end=-1),
                comment=f'REGENERATE (back to stage {action_stage})',
            )
            self._requesting_small_change = False
        else:
            self._requesting_small_change = issues.do_all_issues_request_small_change()
        # add posted issues to the counts:
        for issue in posted_issues:
            self.issues_to_counts[issue] = self.issues_to_counts.get(issue, 0) + 1
        return None

    @_raise_if_reset
    def _get_code_and_respond_to_issues(self, response: str) -> Optional[CodeAndOutput]:
        """
        Get a code from the LLM, run it and return code and result.
        If the code fails, notify the LLM and return None.
        """
        # Try to extract the code:
        code_extractor = self._get_code_extractor()
        try:
            code, num_added_lines = code_extractor.get_modified_code_and_num_added_lines(response)
        except IncompleteBlockFailedExtractingBlock:
            is_bumped = self._bump_to_model_with_more_context()
            return self._respond_to_issues(self._get_issue_for_incomplete_code_block(is_bumped), is_bumped=is_bumped)
        except FailedExtractingBlock as e:
            return self._respond_to_issues(self._get_issue_for_missing_or_multiple_code_blocks(e))

        code_runner_wrapper = self._get_code_runner_wrapper(code)
        code_and_output = self.code_and_output_cls(code=code)
        # We were able to extract the code. We now statically check the code before running it.
        static_code_check_issues = []
        if self._requesting_small_change:
            static_code_check_issues.append(
                self._get_issue_for_new_code_not_being_a_modification_of_old_code(code, self.previous_code))
        static_code_check_issues.extend(code_extractor.get_issues_for_static_code_check(code=code))

        if static_code_check_issues:
            return self._respond_to_issues(static_code_check_issues, code_and_output)

        # Code passes static checks. We can now run the code.
        result, created_files, multi_context, exception = code_runner_wrapper.run()
        issues = multi_context.issues
        contexts = multi_context.contexts
        code_and_output = self._get_code_and_output(code, result, created_files, contexts)

        if exception is not None:
            if isinstance(exception, RunIssue):
                run_time_issue = exception
            elif isinstance(exception, ImportError):
                run_time_issue = self._get_issue_for_allowed_packages(exception.exception, contexts)
            else:
                exceptions_to_funcs = {
                    TimeoutError: self._get_issue_for_timeout,
                    UnAllowedFilesCreated: self._get_issue_for_un_allowed_files_created,
                    FileNotFoundError: self._get_issue_for_file_not_found,
                    CodeUsesForbiddenFunctions: self._get_issue_for_forbidden_functions,
                    CodeImportForbiddenModule: self._get_issue_for_forbidden_import,
                    CodeWriteForbiddenFile: self._get_issue_for_forbidden_write,
                    CodeReadForbiddenFile: self._get_issue_for_forbidden_read,
                }
                for e_type, func in exceptions_to_funcs.items():
                    if isinstance(exception.exception, e_type):
                        run_time_issue = func(exception.exception, exception)
                        break
                else:
                    run_time_issue = self._get_issue_for_regular_exception_or_warning(exception, num_added_lines)
            return self._respond_to_issues(run_time_issue, code_and_output)

        # The code ran without raising exceptions.
        # We now check for issues in the output files as well as issues collected during the run:
        output_issues = []
        output_issues.extend(issues)
        output_issues.extend(self._get_issues_for_created_output_files(code_and_output, contexts))
        output_issues = self._remove_issues_exceeding_max_count(output_issues)
        if output_issues:
            # if the code ran, but output was incorrect, we delete any created files:
            code_and_output.created_files.delete_all_created_files(self.data_folder)
            return self._respond_to_issues(output_issues, code_and_output)

        return self._respond_to_issues(None, code_and_output)

    def run_debugging(self) -> Optional[CodeAndOutput]:
        """
        Run the debugging process.
        If debugging did not converge to a running code within the max_debug_iterations, return None.
        Otherwise, return the code and output.
        """
        self.initialize_conversation_if_needed()
        for self.debug_iteration in range(1, self.max_debug_iterations + 1):
            response = self.apply_get_and_append_assistant_message(is_code=True,
                                                                   previous_code=self.previous_code).content
            with self._app_temporarily_set_panel_status(PanelNames.FEEDBACK, 'Running and checking code'):
                self._app_send_prompt(PanelNames.FEEDBACK)
                code_and_output = self._get_code_and_respond_to_issues(response)
            if code_and_output is not None:
                self._app_send_prompt(
                    PanelNames.FEEDBACK,
                    dedent_triple_quote_str(f"""
                    ## {Symbols.CHECK_SYMBOL} Code check successful
                    Code ran without issues and passed all rule-based checks.
                    You can see code output in the Product panel.
                    """),
                    from_md=True,
                    sleep_for=PAUSE_AT_RULE_BASED_FEEDBACK)
                return code_and_output
        self.apply_append_user_message(
            "It seems like we are not converging. Let's try again from the start.\n"
            "Please provide a fresh new attempt of the code.", ignore=True,
            sleep_for=PAUSE_AT_RULE_BASED_FEEDBACK)
        self._rewind_conversation_to_first_response()

        return None



from copy import copy
from functools import wraps
from typing import Iterable, Dict, Callable, Optional, Tuple, List, Type

import pandas as pd

from pandas.core.frame import DataFrame

from dataclasses import dataclass, field

from pandas.core.indexing import _LocationIndexer

from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.mutable import Flag
from ...base_run_contexts import RunContext
from .dataframe_operations import DataframeOperation, ChangeSeriesDataframeOperation, DataframeOperations, \
    CreationDataframeOperation
from . import df_methods
from ...run_issues import CodeProblem, RunIssue

CLS_METHOD_NAMES_NEW_METHODS = [
    (DataFrame, '__init__', df_methods.__init__),
    (DataFrame, '__setitem__', df_methods.__setitem__),
    (DataFrame, '__getitem__', df_methods.__getitem__),
    (DataFrame, '__delitem__', df_methods.__delitem__),
    (DataFrame, '__str__', df_methods.__str__),
    (DataFrame, 'to_string', df_methods.to_string),
    (DataFrame, 'to_csv', df_methods.to_csv),
    (DataFrame, 'to_latex', df_methods.raise_on_call),
    (DataFrame, 'to_json', df_methods.raise_on_call),
    (_LocationIndexer, '__getitem__', df_methods.__LocationIndexer__get_item__),
]


@dataclass
class DataFrameSeriesChange(RunIssue):
    """
    Exception that is raised when a data frame series is changed.
    """
    changed_series: str = None
    category: str = 'Good coding practices: Dataframe series change'
    issue: str = 'Your code changes the series "{changed_series}" of your dataframe.\n' \
                 'This could lead to confusion and errors.'
    instructions: str = 'Instead of changing an existing dataframe series, please create a new series, and give it a ' \
                        'new sensible name.'
    code_problem: CodeProblem = CodeProblem.RuntimeError
    comment: str = 'Code modifies dataframe series'


@dataclass
class TrackDataFrames(RunContext):
    """
    Context manager that tracks all the data frames that are created and their changes during the context.
    """
    data_frame_operations: DataframeOperations = None

    allow_dataframes_to_change_existing_series: Optional[bool] = True
    # True: allow changing existing series for all df
    # False: raise an exception when trying to change an existing series for all df
    # None: allow changing existing series for all df except for the ones that are created from a file

    enforce_saving_altered_dataframes: bool = False

    str_float_format: str = field(default_factory=lambda: df_methods.STR_FLOAT_FORMAT)

    df_creating_func_names_and_is_file: Iterable[str] = (
        ('read_pickle', True),
        ('read_csv', True),
        ('read_excel', True),
        ('read_json', True),
    )
    cls_method_names_new_methods: List[Tuple[Type, str, Callable]] = \
        field(default_factory=lambda: CLS_METHOD_NAMES_NEW_METHODS)

    _original_float_format: Optional[str] = None
    _df_creating_func_names_to_original_funcs: Optional[Dict[str, Callable]] = None
    _cls_method_names_original_methods: Optional[List[Tuple[Type, str, Callable]]] = None
    _prevent_recording_changes: Flag = field(default_factory=Flag)

    def _df_creating_func_override(self, *args, original_method=None, is_file=False, on_change=None, **kwargs):
        """
        Override for a dataframe creating function.
        Adds a `file_path` and a `created_by` attribute to the created dataframe.
        """
        with self._prevent_recording_changes.temporary_set(True):
            df = original_method(*args, **kwargs)
        if not isinstance(df, pd.DataFrame):
            return df

        if is_file:
            file_path = args[0] if len(args) > 0 else kwargs.get('filepath_or_buffer')
        else:
            file_path = None
        created_by = original_method.__name__
        if isinstance(df, pd.DataFrame):
            df.created_by = created_by
            df.file_path = file_path
            on_change(df, CreationDataframeOperation(
                id=id(df), created_by=created_by, file_path=file_path, columns=copy(df.columns.values)))
        return df

    def _override_df_creating_funcs(self):
        """
        Hook all the dataframe creating functions so that they return a DataFrame with a `file_path` and a
        `created_by` attributes.
        """
        self._df_creating_func_names_to_original_funcs = {}
        for func_name, is_file in self.df_creating_func_names_and_is_file:
            original_func = getattr(pd, func_name)
            assert hasattr(pd, func_name), f"pd does not have a method {func_name}"
            wrapped_new_method = self._get_wrapped_new_method(
                self._df_creating_func_override, original_func, is_file=is_file)
            setattr(pd, func_name, wrapped_new_method)
            self._df_creating_func_names_to_original_funcs[func_name] = original_func

    def _de_override_df_creating_funcs(self):
        """
        De-hook all the dataframe creating functions.
        """
        for func_name, original_func in self._df_creating_func_names_to_original_funcs.items():
            setattr(pd, func_name, original_func)
        self._df_creating_func_names_to_original_funcs = None

    def _get_wrapped_new_method(self, new_method, original_method, **kwargs):
        """
        Wrap a method so that it has the original method as an argument and the `on_change` callback.
        """
        @wraps(original_method)
        def wrapped_new_method(*args, **k):
            return new_method(*args, original_method=original_method, on_change=self._on_change, **k, **kwargs)
        wrapped_new_method.wrapper_of = original_method
        return wrapped_new_method

    def _override_df_methods(self):
        """
        Override specified dataframe methods so that they report changes to the dataframe.
        As well as other enhancements.
        """
        self._cls_method_names_original_methods = []
        for cls, method_name, new_method in self.cls_method_names_new_methods:
            original_method = getattr(cls, method_name)
            wrapped_new_method = self._get_wrapped_new_method(new_method, original_method)
            setattr(cls, method_name, wrapped_new_method)
            self._cls_method_names_original_methods.append((cls, method_name, original_method))

    def _de_override_df_methods(self):
        """
        De-hook all the dataframe methods.
        """
        for cls, method_name, original_method in self._cls_method_names_original_methods:
            setattr(cls, method_name, original_method)
        self._cls_method_names_original_methods = None

    def _override_float_format(self):
        """
        Override the float format.
        """
        if self.str_float_format:
            self._original_float_format = pd.get_option('display.float_format')
            pd.set_option(f'display.float_format', self.str_float_format)
        else:
            self._original_float_format = None

    def _de_override_float_format(self):
        """
        De-hook the float format.
        """
        if self._original_float_format:
            pd.set_option(f'display.float_format', self._original_float_format)
            self._original_float_format = None

    def _on_change(self, df, series_operation: DataframeOperation):
        if self._prevent_recording_changes:
            return
        if isinstance(series_operation, ChangeSeriesDataframeOperation) and self._is_called_from_user_script(5):
            if self.allow_dataframes_to_change_existing_series is False \
                    or (self.allow_dataframes_to_change_existing_series is None and df.file_path is not None):
                self.issues.append(DataFrameSeriesChange.from_current_tb(changed_series=series_operation.series_name))
        self.dataframe_operations.append(series_operation)

    def _create_issues_for_unsaved_dataframes(self):
        if not self.enforce_saving_altered_dataframes:
            return
        dataframe_operations = self.dataframe_operations
        if dataframe_operations.get_read_changed_but_unsaved_ids():
            # Not all changed dataframes were saved to files.
            read_but_unsaved_filenames = dataframe_operations.get_read_filenames_from_ids(
                dataframe_operations.get_read_changed_but_unsaved_ids())
            self.issues.append(RunIssue(
                category='Coding good practice: Any modified dataframe should be saved to a file',
                issue=dedent_triple_quote_str(f"""
                    Your code modifies, but doesn't save, some of the dataframes:
                    {read_but_unsaved_filenames}.
                    """),
                instructions=dedent_triple_quote_str("""
                    The code should use `to_csv` to save any modified dataframe in a new file \t
                    in the same directory as the code.
                    """),
                comment='Not all modified dataframes were saved',
                code_problem=CodeProblem.MissingOutputFiles,
            ))

    def __enter__(self):
        self._override_df_creating_funcs()
        self._override_df_methods()
        self._override_float_format()
        self.dataframe_operations = DataframeOperations()
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._de_override_float_format()
        self._de_override_df_methods()
        self._de_override_df_creating_funcs()
        self._create_issues_for_unsaved_dataframes()
        return super().__exit__(exc_type, exc_val, exc_tb)
